"""
Lyndon words
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import divisors, gcd, moebius, multinomial
from sage.combinat.combinat_cython import lyndon_word_iterator
from sage.combinat.composition import Composition, Compositions
from sage.combinat.necklace import _sfc
from sage.combinat.words.finite_word import FiniteWord_class
from sage.combinat.words.words import FiniteWords
from sage.rings.integer import Integer
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


def LyndonWords(e=None, k=None):
    """
    Return the combinatorial class of Lyndon words.

    A Lyndon word `w` is a word that is lexicographically less than all of
    its rotations.  Equivalently, whenever `w` is split into two non-empty
    substrings, `w` is lexicographically less than the right substring.

    See :wikipedia:`Lyndon_word`

    INPUT:

    - no input at all

    or

    - ``e`` -- integer; size of alphabet
    - ``k`` -- integer; length of the words

    or

    - ``e`` -- a composition

    OUTPUT: a combinatorial class of Lyndon words

    EXAMPLES::

        sage: LyndonWords()
        Lyndon words

    If e is an integer, then e specifies the length of the
    alphabet; k must also be specified in this case::

        sage: LW = LyndonWords(3, 4); LW
        Lyndon words from an alphabet of size 3 of length 4
        sage: LW.first()
        word: 1112
        sage: LW.last()
        word: 2333
        sage: LW.random_element()  # random                                             # needs sage.libs.pari
        word: 1232
        sage: LW.cardinality()                                                          # needs sage.libs.pari
        18

    If e is a (weak) composition, then it returns the class of Lyndon
    words that have evaluation e::

        sage: LyndonWords([2, 0, 1]).list()
        [word: 113]
        sage: LyndonWords([2, 0, 1, 0, 1]).list()
        [word: 1135, word: 1153, word: 1315]
        sage: LyndonWords([2, 1, 1]).list()
        [word: 1123, word: 1132, word: 1213]
    """
    if e is None and k is None:
        return LyndonWords_class()
    if isinstance(e, (int, Integer)):
        if e > 0:
            if not isinstance(k, (int, Integer)):
                raise TypeError("k must be a nonnegative integer")
            if k < 0:
                raise TypeError("k must be a nonnegative integer")
            return LyndonWords_nk(Integer(e), Integer(k))
    elif e in Compositions():
        return LyndonWords_evaluation(Composition(e))

    raise TypeError("e must be a positive integer or a composition")


def LyndonWord(data, check=True):
    r"""
    Construction of a Lyndon word.

    INPUT:

    - ``data`` -- list
    - ``check`` -- boolean (default: ``True``); if ``True``,
      check that the input data represents a Lyndon word

    OUTPUT: a Lyndon word

    EXAMPLES::

        sage: LyndonWord([1,2,2])
        word: 122
        sage: LyndonWord([1,2,3])
        word: 123
        sage: LyndonWord([2,1,2,3])
        Traceback (most recent call last):
        ...
        ValueError: not a Lyndon word

    If ``check`` is ``False``, then no verification is done::

        sage: LyndonWord([2,1,2,3], check=False)
        word: 2123
    """
    return LyndonWords()(data, check=check)


class LyndonWords_class(UniqueRepresentation, Parent):
    r"""
    The set of all Lyndon words.
    """
    def __init__(self, alphabet=None):
        r"""
        INPUT:

        - ``alphabet`` -- the underlying alphabet

        TESTS::

            sage: loads(dumps(LyndonWords())) is LyndonWords()
            True
        """
        from sage.categories.sets_cat import Sets
        self._words = FiniteWords()
        Parent.__init__(self, category=Sets().Infinite(), facade=(self._words))

    def __call__(self, *args, **kwds):
        r"""
        TESTS::

            sage: L = LyndonWords()
            sage: L('aababc')
            word: aababc
            sage: L([2,0,1])
            Traceback (most recent call last):
            ...
            ValueError: not a Lyndon word
        """
        w = self._words(*args, **kwds)
        if kwds.get('check', True) and not w.is_lyndon():
            raise ValueError("not a Lyndon word")
        return w

    def __repr__(self):
        r"""
        String representation.

        EXAMPLES::

            sage: LyndonWords()
            Lyndon words
        """
        return "Lyndon words"

    def __contains__(self, w):
        """
        TESTS::

            sage: LW33 = LyndonWords(3,3)
            sage: all(lw in LyndonWords() for lw in LW33)
            True
        """
        if isinstance(w, list):
            w = self._words(w, check=False)
        return isinstance(w, FiniteWord_class) and w.is_lyndon()


class LyndonWords_evaluation(UniqueRepresentation, Parent):
    r"""
    The set of Lyndon words on a fixed multiset of letters.

    EXAMPLES::

        sage: L = LyndonWords([1,2,1])
        sage: L
        Lyndon words with evaluation [1, 2, 1]
        sage: L.list()
        [word: 1223, word: 1232, word: 1322]
    """
    def __init__(self, e):
        """
        TESTS::

            sage: LW21 = LyndonWords([2,1]); LW21
            Lyndon words with evaluation [2, 1]
            sage: LW21 == loads(dumps(LW21))
            True
        """
        self._e = e
        self._words = FiniteWords(len(e))

        from sage.categories.enumerated_sets import EnumeratedSets
        Parent.__init__(self,
                        category=EnumeratedSets().Finite(),
                        facade=(self._words,)
                        )

    def __repr__(self):
        """
        TESTS::

            sage: repr(LyndonWords([2,1,1]))
            'Lyndon words with evaluation [2, 1, 1]'
        """
        return "Lyndon words with evaluation %s" % self._e

    def __call__(self, *args, **kwds):
        r"""
        TESTS::

            sage: L = LyndonWords([1,2,1])
            sage: L([1,2,2,3])
            word: 1223
            sage: L([2,1,2,3])
            Traceback (most recent call last):
            ...
            ValueError: not a Lyndon word
            sage: L([1,2])
            Traceback (most recent call last):
            ...
            ValueError: evaluation is not [1, 2, 1]
        """
        w = self._words(*args, **kwds)
        if kwds.get('check', True) and not w.is_lyndon():
            raise ValueError("not a Lyndon word")
        if kwds.get('check', True) and w.evaluation() != self._e:
            raise ValueError("evaluation is not {}".format(self._e))
        return w

    def __contains__(self, w):
        """
        EXAMPLES::

            sage: [1,2,1,2] in LyndonWords([2,2])
            False
            sage: [1,1,2,2] in LyndonWords([2,2])
            True
            sage: all(lw in LyndonWords([2,1,3,1]) for lw in LyndonWords([2,1,3,1]))
            True
        """
        if isinstance(w, list):
            w = self._words(w, check=False)
        if isinstance(w, FiniteWord_class) and all(x in self._words.alphabet() for x in w):
            ev_dict = w.evaluation_dict()
            evaluation = [ev_dict.get(x, 0) for x in self._words.alphabet()]
            return evaluation == self._e and w.is_lyndon()
        return False

    def cardinality(self):
        """
        Return the number of Lyndon words with the evaluation e.

        EXAMPLES::

            sage: LyndonWords([]).cardinality()
            0
            sage: LyndonWords([2,2]).cardinality()                                      # needs sage.libs.pari
            1
            sage: LyndonWords([2,3,2]).cardinality()                                    # needs sage.libs.pari
            30

        Check to make sure that the count matches up with the number of
        Lyndon words generated::

            sage: comps = [[],[2,2],[3,2,7],[4,2]] + Compositions(4).list()
            sage: lws = [LyndonWords(comp) for comp in comps]
            sage: all(lw.cardinality() == len(lw.list()) for lw in lws)                 # needs sage.libs.pari
            True
        """
        evaluation = self._e
        le = list(evaluation)
        if not evaluation:
            return Integer(0)
        n = sum(evaluation)
        return sum(moebius(j) * multinomial([ni // j for ni in evaluation])
                   for j in divisors(gcd(le))) // n

    def __iter__(self):
        """
        An iterator for the Lyndon words with evaluation e.

        EXAMPLES::

            sage: LyndonWords([1]).list()    #indirect doctest
            [word: 1]
            sage: LyndonWords([2]).list()    #indirect doctest
            []
            sage: LyndonWords([3]).list()    #indirect doctest
            []
            sage: LyndonWords([3,1]).list()  #indirect doctest
            [word: 1112]
            sage: LyndonWords([2,2]).list()  #indirect doctest
            [word: 1122]
            sage: LyndonWords([1,3]).list()  #indirect doctest
            [word: 1222]
            sage: LyndonWords([3,3]).list()  #indirect doctest
            [word: 111222, word: 112122, word: 112212]
            sage: LyndonWords([4,3]).list()  #indirect doctest
            [word: 1111222, word: 1112122, word: 1112212, word: 1121122, word: 1121212]

        TESTS:

        Check that :issue:`12997` is fixed::

            sage: LyndonWords([0,1]).list()
            [word: 2]
            sage: LyndonWords([0,2]).list()
            []
            sage: LyndonWords([0,0,1,0,1]).list()
            [word: 35]
        """
        if not self._e:
            return
        k = 0
        while self._e[k] == 0:
            k += 1
        for z in _sfc(self._e[k:], equality=True):
            yield self._words([i + k + 1 for i in z], check=False)


class LyndonWords_nk(UniqueRepresentation, Parent):
    r"""
    Lyndon words of fixed length `k` over the alphabet `\{1, 2, \ldots, n\}`.

    INPUT:

    - ``n`` -- the size of the alphabet
    - ``k`` -- the length of the words

    EXAMPLES::

        sage: L = LyndonWords(3, 4)
        sage: L.list()
        [word: 1112,
         word: 1113,
         word: 1122,
         word: 1123,
         ...
         word: 1333,
         word: 2223,
         word: 2233,
         word: 2333]
    """
    def __init__(self, n, k):
        """
        Initialize ``self``.

        TESTS::

            sage: LW23 = LyndonWords(2,3); LW23
            Lyndon words from an alphabet of size 2 of length 3
            sage: LW23== loads(dumps(LW23))
            True
        """
        self._n = n
        self._k = k
        self._words = FiniteWords(self._n)

        from sage.categories.enumerated_sets import EnumeratedSets
        Parent.__init__(self,
                        category=EnumeratedSets().Finite(),
                        facade=(self._words,)
                        )

    def __repr__(self):
        """
        TESTS::

            sage: repr(LyndonWords(2, 3))
            'Lyndon words from an alphabet of size 2 of length 3'
        """
        return "Lyndon words from an alphabet of size %s of length %s" % (self._n, self._k)

    def __call__(self, *args, **kwds):
        r"""
        TESTS::

            sage: L = LyndonWords(3,3)
            sage: L([1,2,3])
            word: 123
            sage: L([2,3,4])
            Traceback (most recent call last):
            ...
            ValueError: 4 not in alphabet
            sage: L([2,1,3])
            Traceback (most recent call last):
            ...
            ValueError: not a Lyndon word
            sage: L([1,2,2,3,3])
            Traceback (most recent call last):
            ...
            ValueError: length is not k=3

        Make sure that the correct length is checked (:issue:`30186`)::

            sage: L = LyndonWords(2, 4)
            sage: _ = L(L.random_element())                                             # needs sage.libs.pari
        """
        w = self._words(*args, **kwds)
        if kwds.get('check', True) and not w.is_lyndon():
            raise ValueError("not a Lyndon word")
        if kwds.get('check', True) and w.length() != self._k:
            raise ValueError("length is not k={}".format(self._k))
        return w

    def __contains__(self, w):
        """
        TESTS::

            sage: LW33 = LyndonWords(3,3)
            sage: all(lw in LW33 for lw in LW33)
            True
        """
        if isinstance(w, list):
            w = self._words(w, check=False)
        return isinstance(w, FiniteWord_class) and w.length() == self._k \
            and all(x in self._words.alphabet() for x in w) and w.is_lyndon()

    def cardinality(self):
        """
        TESTS::

            sage: [ LyndonWords(3,i).cardinality() for i in range(1, 11) ]              # needs sage.libs.pari
            [3, 3, 8, 18, 48, 116, 312, 810, 2184, 5880]
        """
        if self._k == 0:
            return Integer(1)
        s = Integer(0)
        for d in divisors(self._k):
            s += moebius(d) * self._n**(self._k // d)
        return s // self._k

    def __iter__(self):
        """
        TESTS::

            sage: LyndonWords(3,3).list()  # indirect doctest
            [word: 112, word: 113, word: 122, word: 123, word: 132, word: 133, word: 223, word: 233]

            sage: sum(1 for lw in LyndonWords(11, 6))
            295020

            sage: sum(1 for lw in LyndonWords(1000, 1))
            1000

            sage: sum(1 for lw in LyndonWords(1, 1000))
            0

            sage: list(LyndonWords(1, 1))                                               # needs sage.libs.pari
            [word: 1]
        """
        W = self._words._element_classes['list']
        for lw in lyndon_word_iterator(self._n, self._k):
            yield W(self._words, [i + 1 for i in lw])


def StandardBracketedLyndonWords(n, k):
    """
    Return the combinatorial class of standard bracketed Lyndon words
    from [1, ..., n] of length k.

    These are in one to one correspondence with the Lyndon words and
    form a basis for the subspace of degree k of the free Lie algebra
    of rank n.

    EXAMPLES::

        sage: SBLW33 = StandardBracketedLyndonWords(3,3); SBLW33
        Standard bracketed Lyndon words from an alphabet of size 3 of length 3
        sage: SBLW33.first()
        [1, [1, 2]]
        sage: SBLW33.last()
        [[2, 3], 3]
        sage: SBLW33.cardinality()
        8
        sage: SBLW33.random_element() in SBLW33
        True
    """
    return StandardBracketedLyndonWords_nk(n, k)


class StandardBracketedLyndonWords_nk(UniqueRepresentation, Parent):
    def __init__(self, n, k):
        """
        TESTS::

            sage: SBLW = StandardBracketedLyndonWords(3, 2)
            sage: SBLW == loads(dumps(SBLW))
            True
        """
        self._n = n
        self._k = k
        self._lyndon = LyndonWords(self._n, self._k)

        from sage.categories.enumerated_sets import EnumeratedSets
        Parent.__init__(self, category=EnumeratedSets().Finite())

    def __repr__(self):
        """
        TESTS::

            sage: repr(StandardBracketedLyndonWords(3, 3))
            'Standard bracketed Lyndon words from an alphabet of size 3 of length 3'
        """
        return "Standard bracketed Lyndon words from an alphabet of size %s of length %s" % (self._n, self._k)

    def cardinality(self):
        """
        EXAMPLES::

            sage: StandardBracketedLyndonWords(3, 3).cardinality()
            8
            sage: StandardBracketedLyndonWords(3, 4).cardinality()
            18
        """
        return self._lyndon.cardinality()

    def __call__(self, *args, **kwds):
        r"""
        EXAMPLES::

            sage: S = StandardBracketedLyndonWords(3, 3)
            sage: S([1,2,3])
            [1, [2, 3]]
        """
        return standard_bracketing(self._lyndon(*args, **kwds))

    def __contains__(self, sblw):
        """
        EXAMPLES::

            sage: S = StandardBracketedLyndonWords(2, 3)
            sage: [[1, 2], 2] in S
            True
            sage: [1, [2, 2]] in S
            False
            sage: [1, [2, 3]] in S
            False
            sage: [1, 2] in S
            False
        """
        try:
            lw = standard_unbracketing(sblw)
        except ValueError:
            return False
        return len(lw) == self._k and all(a in self._lyndon._words.alphabet() for a in lw.parent().alphabet())

    def __iter__(self):
        """
        EXAMPLES::

            sage: StandardBracketedLyndonWords(3, 3).list()
            [[1, [1, 2]],
             [1, [1, 3]],
             [[1, 2], 2],
             [1, [2, 3]],
             [[1, 3], 2],
             [[1, 3], 3],
             [2, [2, 3]],
             [[2, 3], 3]]
        """
        for x in self._lyndon:
            yield standard_bracketing(x)


def standard_bracketing(lw):
    """
    Return the standard bracketing of a Lyndon word ``lw``.

    EXAMPLES::

        sage: import sage.combinat.words.lyndon_word as lyndon_word
        sage: [lyndon_word.standard_bracketing(u) for u in LyndonWords(3,3)]
        [[1, [1, 2]],
         [1, [1, 3]],
         [[1, 2], 2],
         [1, [2, 3]],
         [[1, 3], 2],
         [[1, 3], 3],
         [2, [2, 3]],
         [[2, 3], 3]]
    """
    if len(lw) == 1:
        return lw[0]

    for i in range(1, len(lw)):
        if lw[i:] in LyndonWords():
            return [standard_bracketing(lw[:i]), standard_bracketing(lw[i:])]


def standard_unbracketing(sblw):
    """
    Return flattened ``sblw`` if it is a standard bracketing of a Lyndon word,
    otherwise raise an error.

    EXAMPLES::

        sage: from sage.combinat.words.lyndon_word import standard_unbracketing
        sage: standard_unbracketing([1, [2, 3]])
        word: 123
        sage: standard_unbracketing([[1, 2], 3])
        Traceback (most recent call last):
        ...
        ValueError: not a standard bracketing of a Lyndon word

    TESTS::

        sage: standard_unbracketing(1) # Letters don't use brackets.
        word: 1
        sage: standard_unbracketing([1])
        Traceback (most recent call last):
        ...
        ValueError: not a standard bracketing of a Lyndon word
    """
    # Nested helper function that not only returns (flattened) w, but also its
    # right factor in the standard Lyndon factorization.
    def standard_unbracketing_rec(w):
        if not isinstance(w, list):
            return [w], []
        if len(w) != 2:
            raise ValueError("not a standard bracketing of a Lyndon word")
        x, t = standard_unbracketing_rec(w[0])
        y, _ = standard_unbracketing_rec(w[1])
        # If x = st is a standard Lyndon factorization, and y is a Lyndon word
        # such that y <= t, then xy is standard (but not necessarily Lyndon).
        if x < y and (len(t) == 0 or y <= t):
            x += y
            return x, y
        raise ValueError("not a standard bracketing of a Lyndon word")
    lw, _ = standard_unbracketing_rec(sblw)
    return FiniteWords(list(set(lw)))(lw, datatype='list', check=False)



r"""
Weierstrass `\wp`-function for elliptic curves

The Weierstrass `\wp` function associated to an elliptic curve over a
field `k` is a Laurent series of the form

.. MATH::

    \wp(z) = \frac{1}{z^2} +  c_2 \cdot z^2 + c_4 \cdot z^4 + \cdots.

If the field is contained in `\mathbb{C}`, then this is the series
expansion of the map from `\mathbb{C}` to `E(\mathbb{C})` whose kernel
is the period lattice of `E`.

Over other fields, like finite fields, this still makes sense as a
formal power series with coefficients in `k` - at least its first `p-2`
coefficients where `p` is the characteristic of `k`. It can be defined
via the formal group as `x+c` in the variable `z=\log_E(t)` for a
constant `c` such that the constant term `c_0` in `\wp(z)` is zero.

EXAMPLES::

    sage: E = EllipticCurve([0,1])
    sage: E.weierstrass_p()
    z^-2 - 1/7*z^4 + 1/637*z^10 - 1/84721*z^16 + O(z^20)

REFERENCES:

- [BMSS2006]_

AUTHORS:

- Dan Shumov 04/09: original implementation

- Chris Wuthrich 11/09: major restructuring

- Jeroen Demeyer (2014-03-06): code clean up, fix characteristic bound
  for quadratic algorithm (see :issue:`15855`)
"""

#*****************************************************************************
#       Copyright (C) 2009 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.rings.laurent_series_ring import LaurentSeriesRing
from sage.rings.power_series_ring import PowerSeriesRing

# Note: Part of the documentation is replicated in ell_field.py for
# users' convenience. Make sure to keep the two copies synchronized.


def weierstrass_p(E, prec=20, algorithm=None):
    r"""
    Compute the Weierstrass `\wp`-function on an elliptic curve.

    INPUT:

    - ``E`` -- an elliptic curve

    - ``prec`` -- precision

    - ``algorithm`` -- string or ``None`` (default: ``None``);
      a choice of algorithm among ``'pari'``, ``'fast'``, ``'quadratic'``,
      or ``None`` to let this function determine the best algorithm to use

    OUTPUT:

    a Laurent series in one variable `z` with coefficients in the base
    field `k` of `E`.

    EXAMPLES::

        sage: E = EllipticCurve('11a1')
        sage: E.weierstrass_p(prec=10)
        z^-2 + 31/15*z^2 + 2501/756*z^4 + 961/675*z^6 + 77531/41580*z^8 + O(z^10)
        sage: E.weierstrass_p(prec=8)
        z^-2 + 31/15*z^2 + 2501/756*z^4 + 961/675*z^6 + O(z^8)
        sage: Esh = E.short_weierstrass_model()
        sage: Esh.weierstrass_p(prec=8)
        z^-2 + 13392/5*z^2 + 1080432/7*z^4 + 59781888/25*z^6 + O(z^8)

        sage: E.weierstrass_p(prec=8, algorithm='pari')
        z^-2 + 31/15*z^2 + 2501/756*z^4 + 961/675*z^6 + O(z^8)
        sage: E.weierstrass_p(prec=8, algorithm='quadratic')
        z^-2 + 31/15*z^2 + 2501/756*z^4 + 961/675*z^6 + O(z^8)

        sage: k = GF(11)
        sage: E = EllipticCurve(k, [1,1])
        sage: E.weierstrass_p(prec=6, algorithm='fast')
        z^-2 + 2*z^2 + 3*z^4 + O(z^6)
        sage: E.weierstrass_p(prec=7, algorithm='fast')
        Traceback (most recent call last):
        ...
        ValueError: for computing the Weierstrass p-function via the fast algorithm,
        the characteristic (11) of the underlying field must be greater than prec + 4 = 11
        sage: E.weierstrass_p(prec=8)
        z^-2 + 2*z^2 + 3*z^4 + 5*z^6 + O(z^8)
        sage: E.weierstrass_p(prec=8, algorithm='quadratic')
        z^-2 + 2*z^2 + 3*z^4 + 5*z^6 + O(z^8)
        sage: E.weierstrass_p(prec=8, algorithm='pari')
        z^-2 + 2*z^2 + 3*z^4 + 5*z^6 + O(z^8)
        sage: E.weierstrass_p(prec=9)
        Traceback (most recent call last):
        ...
        NotImplementedError: currently no algorithms for computing the Weierstrass
        p-function for that characteristic / precision pair is implemented.
        Lower the precision below char(k) - 2
        sage: E.weierstrass_p(prec=9, algorithm='quadratic')
        Traceback (most recent call last):
        ...
        ValueError: for computing the Weierstrass p-function via the quadratic
        algorithm, the characteristic (11) of the underlying field must be greater
        than prec + 2 = 11
        sage: E.weierstrass_p(prec=9, algorithm='pari')
        Traceback (most recent call last):
        ...
        ValueError: for computing the Weierstrass p-function via pari, the
        characteristic (11) of the underlying field must be greater than prec + 2 = 11

    TESTS::

        sage: E.weierstrass_p(prec=4, algorithm='foo')
        Traceback (most recent call last):
        ...
        ValueError: unknown algorithm for computing the Weierstrass p-function
    """
    k = E.base_ring()
    p = k.characteristic()

    # if the algorithm is not set, try to determine algorithm from input
    if algorithm is None:
        if 0 < p <= prec + 2:
            raise NotImplementedError("currently no algorithms for computing the Weierstrass p-function for that characteristic / precision pair is implemented. Lower the precision below char(k) - 2")
        algorithm = "pari"

    if algorithm == "pari":
        if 0 < p <= prec + 2:
            raise ValueError("for computing the Weierstrass p-function via pari, the characteristic (%s) of the underlying field must be greater than prec + 2 = %s" % (p,prec+2))
        return compute_wp_pari(E, prec)

    # quadratic and fast algorithms require short Weierstrass model
    Esh = E.short_weierstrass_model()

    A = Esh.a4()
    B = Esh.a6()

    if algorithm == "quadratic":
        if 0 < p <= prec + 2:
            raise ValueError("for computing the Weierstrass p-function via the quadratic algorithm, the characteristic (%s) of the underlying field must be greater than prec + 2 = %s" % (p,prec+2))
        wp = compute_wp_quadratic(k, A, B, prec)
    elif algorithm == "fast":
        if 0 < p <= prec + 4:
            raise ValueError("for computing the Weierstrass p-function via the fast algorithm, the characteristic (%s) of the underlying field must be greater than prec + 4 = %s" % (p,prec+4))
        wp = compute_wp_fast(k, A, B, prec)
    else:
        raise ValueError("unknown algorithm for computing the Weierstrass p-function")

    R = wp.parent()
    z = R.gen()
    u = E.isomorphism_to(Esh).u
    return wp(z*u) * u**2


def compute_wp_pari(E, prec):
    r"""
    Compute the Weierstrass `\wp`-function with the ``ellwp`` function
    from PARI.

    EXAMPLES::

        sage: E = EllipticCurve([0,1])
        sage: from sage.schemes.elliptic_curves.ell_wp import compute_wp_pari
        sage: compute_wp_pari(E, prec=20)
        z^-2 - 1/7*z^4 + 1/637*z^10 - 1/84721*z^16 + O(z^20)
        sage: compute_wp_pari(E, prec=30)
        z^-2 - 1/7*z^4 + 1/637*z^10 - 1/84721*z^16
        + 3/38548055*z^22 - 4/8364927935*z^28 + O(z^30)
    """
    ep = E.__pari__()
    wpp = ep.ellwp(n=prec)
    k = E.base_ring()
    R = LaurentSeriesRing(k,'z')
    z = R.gen()
    wp = z**(-2)
    for i in range(prec):
        wp += k(wpp[i]) * z**i
    wp = wp.add_bigoh(prec)
    return wp


def compute_wp_quadratic(k, A, B, prec):
    r"""
    Compute the truncated Weierstrass function of an elliptic curve
    defined by short Weierstrass model: `y^2 = x^3 + Ax + B`. Uses an
    algorithm that is of complexity `O(prec^2)`.

    Let p be the characteristic of the underlying field. Then we must
    have either p = 0, or p > prec + 2.

    INPUT:

    - ``k`` -- the field of definition of the curve
    - ``A`` -- and
    - ``B`` -- the coefficients of the elliptic curve
    - ``prec`` -- the precision to which we compute the series

    OUTPUT:

    A Laurent series approximating the Weierstrass `\wp`-function to precision ``prec``.

    ALGORITHM:

    This function uses the algorithm described in section 3.2 of [BMSS2006]_.

    EXAMPLES::

        sage: E = EllipticCurve([7,0])
        sage: E.weierstrass_p(prec=10, algorithm='quadratic')
        z^-2 - 7/5*z^2 + 49/75*z^6 + O(z^10)

        sage: E = EllipticCurve(GF(103), [1,2])
        sage: E.weierstrass_p(algorithm='quadratic')
        z^-2 + 41*z^2 + 88*z^4 + 11*z^6 + 57*z^8 + 55*z^10 + 73*z^12
         + 11*z^14 + 17*z^16 + 50*z^18 + O(z^20)

        sage: from sage.schemes.elliptic_curves.ell_wp import compute_wp_quadratic
        sage: compute_wp_quadratic(E.base_ring(), E.a4(), E.a6(), prec=10)
        z^-2 + 41*z^2 + 88*z^4 + 11*z^6 + 57*z^8 + O(z^10)
    """
    m = (prec + 1)//2
    c = [0 for j in range(m)]
    c[0] = -A/5
    c[1] = -B/7

    # first Z represent z^2
    R = LaurentSeriesRing(k,'z')
    Z = R.gen()
    pe = Z**-1 + c[0]*Z + c[1]*Z**2

    for i in range(3, m):
        t = 0
        for j in range(1, i - 1):
            t += c[j-1]*c[i-2-j]
        ci = (3*t)/((i-2)*(2*i+3))
        pe += ci * Z**i
        c[i-1] = ci

    return pe(Z**2).add_bigoh(prec)


def compute_wp_fast(k, A, B, m):
    r"""
    Compute the Weierstrass function of an elliptic curve defined by short Weierstrass model:
    `y^2 = x^3 + Ax + B`. It does this with as fast as polynomial of degree `m` can be multiplied
    together in the base ring, i.e. `O(M(n))` in the notation of [BMSS2006]_.

    Let `p` be the characteristic of the underlying field: Then we must have either `p=0`, or `p > m + 3`.

    INPUT:

    - ``k`` -- the base field of the curve
    - ``A`` -- and
    - ``B`` -- as the coefficients of the short Weierstrass model `y^2 = x^3 +Ax +B`, and
    - ``m`` -- the precision to which the function is computed to

    OUTPUT: the Weierstrass `\wp` function as a Laurent series to precision `m`

    ALGORITHM:

    This function uses the algorithm described in section 3.3 of
    [BMSS2006]_.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.ell_wp import compute_wp_fast
        sage: compute_wp_fast(QQ, 1, 8, 7)
        z^-2 - 1/5*z^2 - 8/7*z^4 + 1/75*z^6 + O(z^7)

        sage: k = GF(37)
        sage: compute_wp_fast(k, k(1), k(8), 5)
        z^-2 + 22*z^2 + 20*z^4 + O(z^5)
    """
    R = PowerSeriesRing(k,'z',default_prec=m+5)
    z = R.gen()
    s = 2
    f1 = z.add_bigoh(m+3)
    n = 2*m + 4

    # solve the nonlinear differential equation
    while (s < n):
        f1pr = f1.derivative()
        next_s = 2*s - 1

        a = 2*f1pr
        b = -(6*B*(f1**5) + 4*A*(f1**3))
        c = B*(f1**6) + A*f1**4 + 1 - (f1pr**2)

        # we should really be computing only mod z^next_s here.
        # but we loose only a factor 2
        f2 = solve_linear_differential_system(a, b, c, 0)
        # sometimes we get to 0 quicker than s reaches n
        if f2 == 0:
            break
        f1 = f1 + f2
        s = next_s

    R = f1
    Q = R**2
    pe = 1/Q

    return pe


def solve_linear_differential_system(a, b, c, alpha):
    r"""
    Solve a system of linear differential equations: `af' + bf = c` and `f'(0) = \alpha`
    where `a`, `b`, and `c` are power series in one variable and `\alpha` is a constant in the coefficient ring.

    ALGORITHM:

    due to Brent and Kung '78.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.ell_wp import solve_linear_differential_system
        sage: k = GF(17)
        sage: R.<x> = PowerSeriesRing(k)
        sage: a = 1 + x + O(x^7); b = x + O(x^7); c = 1 + x^3 + O(x^7); alpha = k(3)
        sage: f = solve_linear_differential_system(a, b, c, alpha)
        sage: f
        3 + x + 15*x^2 + x^3 + 10*x^5 + 3*x^6 + 13*x^7 + O(x^8)
        sage: a*f.derivative() + b*f - c
        O(x^7)
        sage: f(0) == alpha
        True
    """
    a_recip = 1 / a
    B = b * a_recip
    C = c * a_recip
    int_B = B.integral()
    J = int_B.exp()
    J_recip = 1 / J
    CJ = C * J
    int_CJ = CJ.integral()
    return J_recip * (alpha + int_CJ)



from sage.rings.padics.pow_computer cimport PowComputer_class
from sage.libs.ntl.ntl_ZZ_pContext cimport ntl_ZZ_pContext_class
from sage.libs.ntl.types cimport *
from sage.libs.gmp.types cimport mpz_t

cdef class PowComputer_ext(PowComputer_class):
    cdef ZZ_c* small_powers
    cdef ZZ_c top_power
    cdef ZZ_c temp_z
    cdef long _initialized
    cdef mpz_t temp_m2

    # the following are for unpickling
    cdef object _poly
    cdef object _shift_seed
    cdef object _ext_type

    cdef ZZ_c* pow_ZZ_tmp(self, long n) except NULL
    cdef ZZ_c* pow_ZZ_top(self) noexcept

    cdef void cleanup_ext(self) noexcept

cdef class PowComputer_ZZ_pX(PowComputer_ext):
    cdef ntl_ZZ_pContext_class get_context(self, long n)
    cdef ntl_ZZ_pContext_class get_context_capdiv(self, long n)
    cdef ntl_ZZ_pContext_class get_top_context(self)
    cdef restore_context(self, long n)
    cdef restore_context_capdiv(self, long n)
    cdef void restore_top_context(self) noexcept
    cdef ZZ_pX_Modulus_c* get_modulus(self, long n) noexcept
    cdef ZZ_pX_Modulus_c* get_modulus_capdiv(self, long n) noexcept
    cdef ZZ_pX_Modulus_c* get_top_modulus(self) noexcept
    cdef int eis_shift(self, ZZ_pX_c* x, ZZ_pX_c* a, long n, long finalprec) except -1
    cdef int eis_shift_capdiv(self, ZZ_pX_c* x, ZZ_pX_c* a, long n, long finalprec) except -1
    cdef long capdiv(self, long n) noexcept
    cdef int teichmuller_set_c (self, ZZ_pX_c* x, ZZ_pX_c* a, long absprec) except -1

cdef class PowComputer_ZZ_pX_FM(PowComputer_ZZ_pX):
    cdef ntl_ZZ_pContext_class c
    cdef ZZ_pX_Modulus_c mod

cdef class PowComputer_ZZ_pX_FM_Eis(PowComputer_ZZ_pX_FM):
    cdef int low_length
    cdef int high_length
    cdef ZZ_pX_Multiplier_c* low_shifter
    cdef ZZ_pX_Multiplier_c* high_shifter

    cdef void cleanup_ZZ_pX_FM_Eis(self) noexcept

cdef class PowComputer_ZZ_pX_small(PowComputer_ZZ_pX):
    cdef object c  # using a python list so that we can store ntl_ZZ_pContext_class objects
    cdef ZZ_pX_Modulus_c *mod

    cdef void cleanup_ZZ_pX_small(self) noexcept

cdef class PowComputer_ZZ_pX_small_Eis(PowComputer_ZZ_pX_small):
    cdef int low_length
    cdef int high_length
    cdef ZZ_pX_c* low_shifter
    cdef ZZ_pX_c* high_shifter

    cdef void cleanup_ZZ_pX_small_Eis(self) noexcept

cdef class PowComputer_ZZ_pX_big(PowComputer_ZZ_pX):
    cdef object context_list  # using a python list so that we can store ntl_ZZ_pContext_class objects
    cdef ZZ_pX_Modulus_c *modulus_list

    cdef ntl_ZZ_pContext_class top_context
    cdef ZZ_pX_Modulus_c top_mod

    cdef object context_dict  # currently using a dict, optimize for speed later
    cdef object modulus_dict  # currently using a dict, optimize for speed later

    cdef void cleanup_ZZ_pX_big(self) noexcept

cdef class PowComputer_ZZ_pX_big_Eis(PowComputer_ZZ_pX_big):
    cdef int low_length
    cdef int high_length
    cdef ZZ_pX_c* low_shifter
    cdef ZZ_pX_c* high_shifter

    cdef void cleanup_ZZ_pX_big_Eis(self) noexcept

# cdef class PowComputer_ZZ_pEX(PowComputer_ext):
#    cdef ntl_ZZ_pEContext get_context(self, long n)
#    cdef ntl_ZZ_pEContext get_top_context(self)
#    cdef void restore_context(self, long n)
#    cdef void restore_top_context(self)
#    cdef ZZ_pEX_Modulus_c get_modulus(self, long n)
#    cdef ZZ_pEX_Modulus_c get_top_modulus(self)
#
# cdef class PowComputer_ZZ_pEX_FM(PowComputer_ZZ_pEX):
#    cdef ZZ_pEX_c poly
#    cdef ntl_ZZ_pEContext c
#    cdef ZZ_pEX_Modulus_c mod
#
#
# cdef class PowComputer_ZZ_pEX_small(PowComputer_ZZ_pEX):
#    cdef ZZ_pEX_c *poly
#    cdef ntl_ZZ_pEContext *c
#    cdef ZZ_pEX_Modulus_c *mod
#
# cdef class PowComputer_ZZ_pEX_big(PowComputer_ZZ_pEX):
#    cdef ZZ_pEX_c poly
#    cdef context_dict  # currently using a dict, optimize for speed later
#    cdef modulus_dict  # currently using a dict, optimize for speed later



"""
Conway-Sloane masses
"""
from sage.arith.misc import (fundamental_discriminant,
                             is_prime,
                             kronecker as kronecker_symbol,
                             legendre_symbol,
                             prime_divisors)
from sage.misc.misc_c import prod
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ


def parity(self, allow_rescaling_flag=True):
    r"""
    Return the parity ("even" or "odd") of an integer-valued quadratic
    form over `\ZZ`, defined up to similitude/rescaling of the form so that
    its Jordan component of smallest scale is unimodular.  After this
    rescaling, we say a form is even if it only represents even numbers,
    and odd if it represents some odd number.

    If the ``allow_rescaling_flag`` is set to False, then we require that
    the quadratic form have a Gram matrix with coefficients in `\ZZ`, and
    look at the unimodular Jordan block to determine its parity.  This
    returns an error if the form is not integer-matrix, meaning that it
    has Jordan components at `p=2` which do not have an integer scale.

    We determine the parity by looking for a `1 \times 1` block in the 0-th
    Jordan component, after a possible rescaling.

    INPUT:

    - ``self`` -- a quadratic form with base ring `\ZZ`, which we may
      require to have integer Gram matrix

    OUTPUT: one of the strings: ``'even'`` or ``'odd'``

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [4, -2, 0, 2, 3, 2]); Q
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 4 -2 0 ]
        [ * 2 3 ]
        [ * * 2 ]
        sage: Q.parity()
        'even'

    ::

        sage: Q = QuadraticForm(ZZ, 3, [4, -2, 0, 2, 3, 1]); Q
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 4 -2 0 ]
        [ * 2 3 ]
        [ * * 1 ]
        sage: Q.parity()
        'even'

    ::

        sage: Q = QuadraticForm(ZZ, 3, [4, -2, 0, 2, 2, 2]); Q
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 4 -2 0 ]
        [ * 2 2 ]
        [ * * 2 ]
        sage: Q.parity()
        'even'

    ::

        sage: Q = QuadraticForm(ZZ, 3, [4, -2, 0, 2, 2, 1]); Q
        Quadratic form in 3 variables over Integer Ring with coefficients:
        [ 4 -2 0 ]
        [ * 2 2 ]
        [ * * 1 ]
        sage: Q.parity()
        'odd'
    """
    # Deal with 0-dim'l forms
    if self.dim() == 0:
        return "even"

    # Identify the correct Jordan component to use.
    Jordan_list = self.jordan_blocks_by_scale_and_unimodular(2)
    scale_pow_list = [J[0] for J in Jordan_list]
    min_scale_pow = min(scale_pow_list)
    if allow_rescaling_flag:
        ind = scale_pow_list.index(min_scale_pow)
    else:
        if min_scale_pow < 0:
            raise TypeError("if rescaling is not allowed, then we require our form to have an integral Gram matrix")
        ind = scale_pow_list.index(0)

    # Find the component of scale (power) zero, and then look for an odd dim'l component.
    J0 = Jordan_list[ind]
    Q0 = J0[1]

    # The lattice is even if there is no component of scale (power) 0
    if J0 is None:
        return "even"

    # Look for a 1x1 block in the 0-th Jordan component (which by
    # convention of the local_normal_form routine will appear first).
    if Q0.dim() == 1:
        return "odd"
    if Q0[0, 1] == 0:
        return "odd"
    return "even"


def is_even(self, allow_rescaling_flag=True) -> bool:
    r"""
    Return true iff after rescaling by some appropriate factor, the
    form represents no odd integers.  For more details, see :meth:`parity`.

    Requires that `Q` is defined over `\ZZ`.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [1, 0, 1])
        sage: Q.is_even()
        False
        sage: Q = QuadraticForm(ZZ, 2, [1, 1, 1])
        sage: Q.is_even()
        True
    """
    return self.parity(allow_rescaling_flag) == "even"


def is_odd(self, allow_rescaling_flag=True) -> bool:
    r"""
    Return true iff after rescaling by some appropriate factor, the
    form represents some odd integers.  For more details, see :meth:`parity`.

    Requires that `Q` is defined over `\ZZ`.

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 2, [1, 0, 1])
        sage: Q.is_odd()
        True
        sage: Q = QuadraticForm(ZZ, 2, [1, 1, 1])
        sage: Q.is_odd()
        False
    """
    return self.parity(allow_rescaling_flag) == "odd"


def conway_species_list_at_odd_prime(self, p):
    r"""
    Return an integer called the 'species' which determines the type
    of the orthogonal group over the finite field `\GF{p}`.

    This assumes that the given quadratic form is a unimodular Jordan
    block at an odd prime `p`.  When the dimension is odd then this
    number is always positive, otherwise it may be positive or
    negative (or zero, but that is considered positive by convention).

    .. NOTE::

        The species of a zero dimensional form is always 0+, so we
        interpret the return value of zero as positive here! =)

    INPUT:

    - ``p`` -- a positive prime number

    OUTPUT: list of integers

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,10))
        sage: Q.conway_species_list_at_odd_prime(3)
        [6, 2, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,8))
        sage: Q.conway_species_list_at_odd_prime(3)
        [5, 2]
        sage: Q.conway_species_list_at_odd_prime(5)
        [-6, 1]
    """
    # Sanity Check:
    if not (p > 2 and is_prime(p)):
        raise TypeError("we are assuming that p is an odd positive prime number")

    # Deal with the zero-dim'l form
    if self.dim() == 0:
        return [0]

    # List the (unscaled/unimodular) Jordan blocks by their scale power
    jordan_list = self.jordan_blocks_in_unimodular_list_by_scale_power(p)

    # Make a list of species (including the two zero-dim'l forms missing at either end of the list of Jordan blocks)
    species_list = []
    for tmp_Q in jordan_list:

        # Some useful variables
        n = tmp_Q.dim()
        d = tmp_Q.det()

        # Determine the species
        if n % 2 != 0:                          # Deal with odd dim'l forms
            species = n
        elif n % 4 == 2 and p % 4 == 3:         # Deal with even dim'l forms
            species = (-1) * legendre_symbol(d, p) * n
        else:
            species = legendre_symbol(d, p) * n

        # Append the species to the list
        species_list.append(species)

    # Return the species list
    return species_list


def conway_species_list_at_2(self):
    r"""
    Return an integer called the 'species' which determines the type
    of the orthogonal group over the finite field `\GF{p}`.

    This assumes that the given quadratic form is a unimodular Jordan
    block at an odd prime `p`.  When the dimension is odd then this
    number is always positive, otherwise it may be positive or
    negative.

    .. NOTE::

        The species of a zero dimensional form is always 0+, so we
        interpret the return value of zero as positive here! =)

    OUTPUT: list of integers

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,10))
        sage: Q.conway_species_list_at_2()
        [1, 5, 1, 1, 1, 1]

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,8))
        sage: Q.conway_species_list_at_2()
        [1, 3, 1, 1, 1]
    """
    # Some useful variables
    n = self.dim()
    d = self.det()

    # Deal with the zero-dim'l form
    if n == 0:
        return 0

    # List the (unscaled/unimodular) Jordan blocks by their scale power
    jordan_list = self.jordan_blocks_in_unimodular_list_by_scale_power(2)

    # Make a list of species (including the two zero-dim'l forms missing at either end of the list of Jordan blocks)
    species_list = []

    if jordan_list[0].parity() == "odd":        # Add an entry for the unlisted "-1" Jordan component as well.
        species_list.append(1)

    for i in range(len(jordan_list)):           # Add an entry for each (listed) Jordan component

        # Make the number 2*t in the C-S Table 1.
        d = jordan_list[i].dim()
        if jordan_list[i].is_even():
            two_t = d
        else:
            two_t = ZZ(2) * ((d - 1) // 2)

        # Determine if the form is bound
        if len(jordan_list) == 1:
            is_bound = False
        elif i == 0:
            is_bound = jordan_list[i + 1].is_odd()
        elif i == len(jordan_list) - 1:
            is_bound = jordan_list[i - 1].is_odd()
        else:
            is_bound = jordan_list[i - 1].is_odd() or jordan_list[i + 1].is_odd()

        # Determine the species
        octane = jordan_list[i].conway_octane_of_this_unimodular_Jordan_block_at_2()
        if is_bound or (octane == 2) or (octane == 6):
            species = two_t + 1
        elif (octane == 0) or (octane == 1) or (octane == 7):
            species = two_t
        else:
            species = (-1) * two_t

        # Append the species to the list
        species_list.append(species)

    if jordan_list[-1].is_odd():        # Add an entry for the unlisted "s_max + 1" Jordan component as well.
        species_list.append(1)

    # Return the species list
    return species_list


def conway_octane_of_this_unimodular_Jordan_block_at_2(self):
    r"""
    Determines the 'octane' of this full unimodular Jordan block at
    the prime `p=2`.  This is an invariant defined (mod 8), ad.

    This assumes that the form is given as a block diagonal form with
    unimodular blocks of size `\leq 2` and the `1 \times 1` blocks are all in the upper
    leftmost position.

    OUTPUT: integer `0 \leq x \leq 7`

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,3,5,7])
        sage: Q.conway_octane_of_this_unimodular_Jordan_block_at_2()
        0
        sage: Q = DiagonalQuadraticForm(ZZ, [1,5,13])
        sage: Q.conway_octane_of_this_unimodular_Jordan_block_at_2()
        3
        sage: Q = DiagonalQuadraticForm(ZZ, [3,7,13])
        sage: Q.conway_octane_of_this_unimodular_Jordan_block_at_2()
        7
    """
    # Deal with 'even' forms
    if self.parity() == "even":
        d = self.Gram_matrix().det()
        if d % 8 == 1 or d % 8 == 7:
            return 0
        return 4

    # Deal with 'odd' forms by diagonalizing, and then computing the octane.
    n = self.dim()
    u = self[0, 0]
    tmp_diag_vec = [None] * n
    tmp_diag_vec[0] = u       # This should be an odd integer!
    ind = 1                   # The next index to diagonalize

    # Use u to diagonalize the form -- WHAT ARE THE POSSIBLE LOCAL NORMAL FORMS?
    while ind < n:

        # Check for a 1x1 block and diagonalize it
        if ind == n - 1 or self[ind, ind + 1] == 0:
            tmp_diag_vec[ind] = self[ind, ind]
            ind += 1

        # Diagonalize the 2x2 block
        else:
            B = self[ind, ind + 1]
            if B % 2:
                raise RuntimeError("we expected the mixed term to be even")

            a = self[ind, ind]
            b = B // ZZ(2)
            c = self[ind + 1, ind + 1]
            tmp_disc = b * b - a * c

            # Perform the diagonalization
            if tmp_disc % 8 == 1:                # 2xy
                tmp_diag_vec[ind] = 1
                tmp_diag_vec[ind + 1] = -1
                ind += 2
            elif tmp_disc % 8 == 5:               # 2x^2 + 2xy + 2y^2
                tmp_diag_vec[0] = 3 * u
                tmp_diag_vec[ind] = -u
                tmp_diag_vec[ind + 1] = -u
                ind += 2
                u = tmp_diag_vec[0]
            else:
                raise RuntimeError("this should not happen -- the odd 2x2 blocks have disc 1 or 5 (mod 8)")

    # Compute the octane
    octane = 0
    for a in tmp_diag_vec:
        if a % 4 == 1:
            octane += 1
        elif a % 4 == 3:
            octane += -1
        else:
            raise RuntimeError("the diagonal elements should all be odd")

    # Return its value
    return octane % 8


def conway_diagonal_factor(self, p):
    r"""
    Compute the diagonal factor of Conway's `p`-mass.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT: a rational number > 0

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,6))
        sage: Q.conway_diagonal_factor(3)
        81/256
    """
    # Get the species list at p
    if p == 2:
        species_list = self.conway_species_list_at_2()
    else:
        species_list = self.conway_species_list_at_odd_prime(p)

    # Evaluate the diagonal factor
    diag_factor = QQ(1)
    for s in species_list:
        if s == 0:
            pass
        elif s % 2 == 1:                   # Note: Here always s > 0.
            diag_factor = diag_factor / (2 * prod([1 - QQ(p)**(-i)
                                                   for i in range(2, s, 2)]))
        else:
            diag_factor = diag_factor / (2 * prod([1 - QQ(p)**(-i)
                                                   for i in range(2, abs(s), 2)]))
            s_sign = ZZ(s / abs(s))
            diag_factor = diag_factor / (ZZ(1) - s_sign * QQ(p) ** ZZ(-abs(s) / ZZ(2)))

    # Return the diagonal factor
    return diag_factor


def conway_cross_product_doubled_power(self, p):
    r"""
    Compute twice the power of `p` which evaluates the 'cross product'
    term in Conway's mass formula.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,8))
        sage: Q.conway_cross_product_doubled_power(2)
        18
        sage: Q.conway_cross_product_doubled_power(3)
        10
        sage: Q.conway_cross_product_doubled_power(5)
        6
        sage: Q.conway_cross_product_doubled_power(7)
        6
        sage: Q.conway_cross_product_doubled_power(11)
        0
        sage: Q.conway_cross_product_doubled_power(13)
        0
    """
    dim_list = [J.dim() for J in self.jordan_blocks_in_unimodular_list_by_scale_power(p)]
    return sum((i - j) * dimi * dim_list[j]
               for i, dimi in enumerate(dim_list)
               for j in range(i))


def conway_type_factor(self):
    r"""
    This is a special factor only present in the mass formula when `p=2`.

    OUTPUT: a rational number

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1,8))
        sage: Q.conway_type_factor()
        4
    """
    jordan_list = self.jordan_blocks_in_unimodular_list_by_scale_power(2)
    n2 = sum([J.dim() for J in jordan_list if J.is_even()])
    n11 = sum([1 for i in range(len(jordan_list) - 1)
               if jordan_list[i].is_odd() and jordan_list[i + 1].is_odd()])

    return ZZ(2)**(n11 - n2)


def conway_p_mass(self, p):
    r"""
    Compute Conway's `p`-mass.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT: a rational number > 0

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, range(1, 6))
        sage: Q.conway_p_mass(2)
        16/3
        sage: Q.conway_p_mass(3)
        729/256
    """
    # Compute the first two factors of the p-mass
    p_mass = self.conway_diagonal_factor(p) * (p ** (self.conway_cross_product_doubled_power(p) / ZZ(2)))

    # Multiply by the 'type factor' when p = 2
    if p == 2:
        p_mass *= self.conway_type_factor()

    # Return the result
    return p_mass


def conway_standard_p_mass(self, p):
    r"""
    Compute the standard (generic) Conway-Sloane `p`-mass.

    INPUT:

    - ``p`` -- a prime number > 0

    OUTPUT: a rational number > 0

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.conway_standard_p_mass(2)
        2/3
    """
    # Some useful variables
    n = self.dim()
    if n % 2 == 0:
        s = n // 2
    else:
        s = (n + 1) // 2

    # Compute the inverse of the generic p-mass
    p_mass_inv = 2 * prod([1 - p**(-i) for i in range(2, 2 * s, 2)])
    if n % 2 == 0:
        D = (-1)**s * self.det() * (2**n)
        #   We should have something like  D = (-1)**s * self.det() / (2**n), but that's not an integer and here we only care about the square-class.
        # d = self.det()   # Note: No normalizing power of 2 is needed since the power is even.
        # if not ((p == 2) or (d % p == 0)):
        p_mass_inv *= (1 - kronecker_symbol(fundamental_discriminant(D), p) * p**(-s))

    # Return the standard p-mass
    return ZZ.one() / p_mass_inv


def conway_standard_mass(self):
    r"""
    Return the infinite product of the standard mass factors.

    OUTPUT: a rational number > 0

    EXAMPLES::

        sage: Q = QuadraticForm(ZZ, 3, [2, -2, 0, 3, -5, 4])
        sage: Q.conway_standard_mass()                                                  # needs sage.symbolic
        1/6

    ::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.conway_standard_mass()                                                  # needs sage.symbolic
        1/6
    """
    from sage.quadratic_forms.special_values import gamma__exact, zeta__exact, quadratic_L_function__exact
    from sage.symbolic.constants import pi

    n = self.dim()
    if n % 2 == 0:
        s = n // 2
    else:
        s = (n + 1) // 2

    generic_mass = 2 * pi**((-1) * n * (n + 1) / ZZ(4)) \
        * prod([gamma__exact(j / ZZ(2)) for j in range(1, n + 1)]) \
        * prod([zeta__exact(2 * k) for k in range(1, s)])

    if n % 2 == 0:
        D = (-1)**s * self.det() * (2**n)
        # We should have something like D = (-1)**s * self.det() / (2**n), but
        # that's not an integer and here we only care about the square-class.
        generic_mass *= quadratic_L_function__exact(s, D)

    return generic_mass


def conway_mass(self):
    r"""
    Compute the mass by using the Conway-Sloane mass formula.

    OUTPUT: a rational number > 0

    EXAMPLES::

        sage: Q = DiagonalQuadraticForm(ZZ, [1,1,1])
        sage: Q.conway_mass()                                                           # needs sage.symbolic
        1/48

        sage: Q = DiagonalQuadraticForm(ZZ, [7,1,1])
        sage: Q.conway_mass()                                                           # needs sage.symbolic
        3/16

        sage: Q = QuadraticForm(ZZ, 3, [7, 2, 2, 2, 0, 2]) + DiagonalQuadraticForm(ZZ, [1])
        sage: Q.conway_mass()                                                           # needs sage.symbolic
        3/32

        sage: Q = QuadraticForm(Matrix(ZZ, 2, [2,1,1,2]))
        sage: Q.conway_mass()                                                           # needs sage.symbolic
        1/12
    """
    # Try to use the cached result
    try:
        return self.__conway_mass
    except AttributeError:
        # Double the form so it's integer-matrix
        Q = self.scale_by_factor(2)

        # Compute the standard mass
        mass = Q.conway_standard_mass()

        # Adjust the p-masses when p|2d
        d = self.det()
        for p in prime_divisors(2 * d):
            mass *= (Q.conway_p_mass(p) / Q.conway_standard_p_mass(p))

        # Cache and return the (simplified) result
        self.__conway_mass = QQ(mass.canonicalize_radical()).abs()
        return self.__conway_mass


# def conway_generic_mass(self):
#    """
#    Computes the generic mass given as
#         2 \pi^{-n(n+1)/4} \prod_{j=1}^{n} \Gamma\(\tfrac{j}{2}\)
#        \zeta(2) \cdots \zeta(2s-2) \zeta_{D}(s)
#    where $n = 2s$ or $2s-1$ depending on the parity of $n$,
#    and $D = (-1)^{s} d$.  We interpret the symbol $\(\frac{D}{p}\)$
#    as 0 if $p\mid 2d$.
#    (Conway and Sloane, Mass formula paper, p??)
#
#    This is possibly equal to
#        2^{-td} * \tau(G) *[\prod_{i=1}^{t} \zeta(1-2i) ]* L(1-t, \chi)
#    where $\dim(Q) = n = 2t$ or $2t+1$, and the last factor is omitted
#    when $n$ is odd.
#    (GHY, Prop 7.4 and 7.5, p121)
#    """
#    RR = RealField(200)
#    n = self.dim()
#    if n % 2 == 0:
#        s = n / 2
#    else:
#        s = (n-1) / 2
#
#    # Form the generic zeta product
#    ans = 2 * RR(pi)^(-n * (n+1) / 4)
#    for j in range(1,n+1):
#        ans *= gamma(RR(j/2))
#    for j in range(2, 2*s, 2):  # j = 2, ..., 2s-2
#        ans *= zeta(RR(j))
#
#    # Extra L-factor for even dimensional forms  -- DO THIS!!!
#    raise NotImplementedError("This routine is not finished yet")
#
#    return ans



# distutils: libraries = GSL_LIBRARIES
# distutils: library_dirs = GSL_LIBDIR
# distutils: include_dirs = GSL_INCDIR
cdef extern from "gsl/gsl_odeiv.h":

  ctypedef struct gsl_odeiv_system:
    int (* function) (double t,  double y[], double dydt[], void * params)
    int (* jacobian) (double t,  double y[], double * dfdy, double dfdt[], void * params)
    size_t dimension
    void * params

  # no:
  #define GSL_ODEIV_FN_EVAL(S,t,y,f)  (*((S)->function))(t,y,f,(S)->params)
  #define GSL_ODEIV_JA_EVAL(S,t,y,dfdy,dfdt)  (*((S)->jacobian))(t,y,dfdy,dfdt,(S)->params)


  ctypedef struct gsl_odeiv_step_type

  ctypedef struct gsl_odeiv_step

  gsl_odeiv_step_type *gsl_odeiv_step_rk2
  gsl_odeiv_step_type *gsl_odeiv_step_rk4
  gsl_odeiv_step_type *gsl_odeiv_step_rkf45
  gsl_odeiv_step_type *gsl_odeiv_step_rkck
  gsl_odeiv_step_type *gsl_odeiv_step_rk8pd
  gsl_odeiv_step_type *gsl_odeiv_step_rk2imp
  gsl_odeiv_step_type *gsl_odeiv_step_rk4imp
  gsl_odeiv_step_type *gsl_odeiv_step_bsimp
  gsl_odeiv_step_type *gsl_odeiv_step_gear1
  gsl_odeiv_step_type *gsl_odeiv_step_gear2


  gsl_odeiv_step * gsl_odeiv_step_alloc( gsl_odeiv_step_type * T, size_t dim)
  int  gsl_odeiv_step_reset(gsl_odeiv_step * s)
  void gsl_odeiv_step_free(gsl_odeiv_step * s)

  char * gsl_odeiv_step_name( gsl_odeiv_step *)
  unsigned int gsl_odeiv_step_order( gsl_odeiv_step * s)

  int  gsl_odeiv_step_apply(gsl_odeiv_step *, double t, double h, double y[], double yerr[],  double dydt_in[], double dydt_out[],  gsl_odeiv_system * dydt)


  ctypedef struct gsl_odeiv_control_type

  ctypedef struct gsl_odeiv_control
  cdef enum:
    GSL_ODEIV_HADJ_DEC = -1
    GSL_ODEIV_HADJ_NIL = 0
    GSL_ODEIV_HADJ_INC = 1

  gsl_odeiv_control * gsl_odeiv_control_alloc( gsl_odeiv_control_type * T)
  int gsl_odeiv_control_init(gsl_odeiv_control * c, double eps_abs, double eps_rel, double a_y, double a_dydt)
  void gsl_odeiv_control_free(gsl_odeiv_control * c)
  int gsl_odeiv_control_hadjust (gsl_odeiv_control * c, gsl_odeiv_step * s,  double y0[],  double yerr[],  double dydt[], double * h)
  char * gsl_odeiv_control_name( gsl_odeiv_control * c)


  gsl_odeiv_control * gsl_odeiv_control_standard_new(double eps_abs, double eps_rel, double a_y, double a_dydt)
  gsl_odeiv_control * gsl_odeiv_control_y_new(double eps_abs, double eps_rel)
  gsl_odeiv_control * gsl_odeiv_control_yp_new(double eps_abs, double eps_rel)

  gsl_odeiv_control * gsl_odeiv_control_scaled_new(double eps_abs, double eps_rel, double a_y, double a_dydt,  double scale_abs[], size_t dim)

  ctypedef struct gsl_odeiv_evolve

  gsl_odeiv_evolve * gsl_odeiv_evolve_alloc(size_t dim)
  int gsl_odeiv_evolve_apply(gsl_odeiv_evolve *, gsl_odeiv_control * con, gsl_odeiv_step * step,  gsl_odeiv_system * dydt, double * t, double t1, double * h, double y[])
  int gsl_odeiv_evolve_reset(gsl_odeiv_evolve *)
  void gsl_odeiv_evolve_free(gsl_odeiv_evolve *)



"""
Heisenberg Algebras

AUTHORS:

- Travis Scrimshaw (2013-08-13): Initial version
"""

# ***************************************************************************
#       Copyright (C) 2013-2017 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************

from sage.misc.cachefunc import cached_method
from sage.structure.indexed_generators import IndexedGenerators

from sage.algebras.lie_algebras.lie_algebra import (LieAlgebraFromAssociative,
                                                    LieAlgebraWithGenerators)
from sage.algebras.lie_algebras.lie_algebra_element import (LieAlgebraElement,
                                                            LieAlgebraMatrixWrapper)
from sage.categories.lie_algebras import LieAlgebras
from sage.categories.cartesian_product import cartesian_product
from sage.matrix.matrix_space import MatrixSpace
from sage.rings.integer import Integer
from sage.sets.disjoint_union_enumerated_sets import DisjointUnionEnumeratedSets
from sage.sets.family import Family
from sage.sets.positive_integers import PositiveIntegers
from sage.sets.set import Set


class HeisenbergAlgebra_abstract(IndexedGenerators):
    """
    The common methods for the (non-matrix) Heisenberg algebras.
    """
    def __init__(self, I):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo) # indirect doctest
        """
        IndexedGenerators.__init__(self, I, prefix='', bracket=False,
                                   latex_bracket=False, string_quotes=False)

    def p(self, i):
        """
        The generator `p_i` of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L.p(2)
            p2
        """
        return self.element_class(self, {'p%i' % i: self.base_ring().one()})

    def q(self, i):
        """
        The generator `q_i` of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L.q(2)
            q2
        """
        return self.element_class(self, {'q%i' % i: self.base_ring().one()})

    def z(self):
        """
        Return the basis element `z` of the Heisenberg algebra.

        The element `z` spans the center of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L.z()
            z
        """
        return self.element_class(self, {'z': self.base_ring().one()})

    def bracket_on_basis(self, x, y):
        """
        Return the bracket of basis elements indexed by ``x`` and ``y``
        where ``x < y``.

        The basis of a Heisenberg algebra is ordered in such a way that
        the `p_i` come first, the `q_i` come next, and the `z` comes last.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 3)
            sage: p1 = ('p', 1)
            sage: q1 = ('q', 1)
            sage: H.bracket_on_basis(p1, q1)
            z
        """
        if y == 'z': # No need to test for x == 'z' since x < y is assumed.
            return self.zero()
        if x[0] == 'p' and y[0] == 'q' and x[1] == y[1]:
            return self.z()
        return self.zero()

    def _repr_term(self, m):
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 3)
            sage: H._repr_term('p1')
            'p1'
            sage: H._repr_term('z')
            'z'
        """
        return m

    def _ascii_art_term(self, m):
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 3)
            sage: H._ascii_art_term('p1')
            p1
            sage: H._ascii_art_term('z')
            z
            sage: ascii_art(sum(i * b for i, b in enumerate(H.basis())))
            p2 + 2*p3 + 3*q1 + 4*q2 + 5*q3 + 6*z
        """
        from sage.typeset.ascii_art import ascii_art
        return ascii_art(m)

    def _latex_term(self, m):
        r"""
        Return a string representation of the term indexed by ``m``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 10)
            sage: H._latex_term('p1')
            'p_{1}'
            sage: H._latex_term('z')
            'z'
            sage: latex(H.p(10))
            p_{10}
        """
        if len(m) == 1:
            return m
        return "%s_{%s}" % (m[0], m[1:]) # else it is of length at least 2

    def _unicode_art_term(self, m):
        r"""
        Return a unicode art representation of the term indexed by ``m``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 10)
            sage: H._unicode_art_term('p1')
            p₁
            sage: H._unicode_art_term('z')
            z
            sage: unicode_art(H.p(10))
            p₁₀
        """
        from sage.typeset.unicode_art import unicode_art, unicode_subscript
        if len(m) == 1:
            return unicode_art(m)
        return unicode_art(str(m[0]) + unicode_subscript(m[1:])) # else it is of length at least 2

    def step(self):
        r"""
        Return the nilpotency step of ``self``.

        EXAMPLES::

            sage: h = lie_algebras.Heisenberg(ZZ, 10)
            sage: h.step()
            2

            sage: h = lie_algebras.Heisenberg(ZZ, oo)
            sage: h.step()
            2
        """
        return Integer(2)

    class Element(LieAlgebraElement):
        pass


class HeisenbergAlgebra_fd:
    """
    Common methods for finite-dimensional Heisenberg algebras.
    """
    def __init__(self, n):
        """
        Initialize ``self``.

        INPUT:

        - ``n`` -- the rank

        TESTS::

            sage: H = lie_algebras.Heisenberg(QQ, 3) # indirect doctest
        """
        self._n = n

    def n(self):
        """
        Return the rank of the Heisenberg algebra ``self``.

        This is the ``n`` such that ``self`` is the `n`-th Heisenberg
        algebra. The dimension of this Heisenberg algebra is then
        `2n + 1`.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 3)
            sage: H.n()
            3
            sage: H = lie_algebras.Heisenberg(QQ, 3, representation='matrix')
            sage: H.n()
            3
        """
        return self._n

    @cached_method
    def gens(self) -> tuple:
        """
        Return the Lie algebra generators of ``self``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 2)
            sage: H.gens()
            (p1, p2, q1, q2)
            sage: H = lie_algebras.Heisenberg(QQ, 0)
            sage: H.gens()
            (z,)
        """
        return tuple(self.lie_algebra_generators())

    def gen(self, i):
        """
        Return the ``i``-th generator of ``self``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 2)
            sage: H.gen(0)
            p1
            sage: H.gen(3)
            q2
        """
        return self.gens()[i]

    @cached_method
    def lie_algebra_generators(self):
        """
        Return the Lie algebra generators of ``self``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 1)
            sage: H.lie_algebra_generators()
            Finite family {'p1': p1, 'q1': q1}
            sage: H = lie_algebras.Heisenberg(QQ, 0)
            sage: H.lie_algebra_generators()
            Finite family {'z': z}
        """
        if self._n == 0:
            return Family(['z'], lambda i: self.z())
        k = ['p%s' % i for i in range(1, self._n+1)]
        k += ['q%s' % i for i in range(1, self._n+1)]
        d = {}
        for i in range(1, self._n+1):
            d['p%s' % i] = self.p(i)
            d['q%s' % i] = self.q(i)
        return Family(k, lambda i: d[i])

    @cached_method
    def basis(self):
        """
        Return the basis of ``self``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, 1)
            sage: H.basis()
            Finite family {'p1': p1, 'q1': q1, 'z': z}
        """
        d = {}
        for i in range(1, self._n+1):
            d['p%s' % i] = self.p(i)
            d['q%s' % i] = self.q(i)
        d['z'] = self.z()
        return Family(self._indices, lambda i: d[i])

    def _coerce_map_from_(self, H):
        """
        Return the coercion map from ``H`` to ``self`` if one exists,
        otherwise return ``None``.

        EXAMPLES::

            sage: HB = lie_algebras.Heisenberg(QQ, 3)
            sage: HM = lie_algebras.Heisenberg(QQ, 3, representation='matrix')
            sage: HB.has_coerce_map_from(HM)
            True
            sage: HM.has_coerce_map_from(HB)
            True
            sage: HB(HM.p(2))
            p2
            sage: HM(-HB.q(3)) == -HM.q(3)
            True
            sage: HB(HM.z())
            z
            sage: HM(HB.z()) == HM.z()
            True
            sage: HQ = lie_algebras.Heisenberg(QQ, 2)
            sage: HB.has_coerce_map_from(HQ)
            True
            sage: HB(HQ.p(2))
            p2
            sage: HZ = lie_algebras.Heisenberg(ZZ, 2)
            sage: HB.has_coerce_map_from(HZ)
            True
            sage: HB(HZ.p(2))
            p2
            sage: HZ = lie_algebras.Heisenberg(ZZ, 2, representation='matrix')
            sage: HB.has_coerce_map_from(HZ)
            True
            sage: HB(HZ.p(2))
            p2
        """
        if isinstance(H, HeisenbergAlgebra_fd):
            if H._n <= self._n and self.base_ring().has_coerce_map_from(H.base_ring()):
                return H.module_morphism(lambda i: self.basis()[i], codomain=self)
            return None # Otherwise no coercion
        return super()._coerce_map_from_(H)


class HeisenbergAlgebra(HeisenbergAlgebra_fd, HeisenbergAlgebra_abstract,
                        LieAlgebraWithGenerators):
    r"""
    A Heisenberg algebra defined using structure coefficients.

    The `n`-th Heisenberg algebra (where `n` is a nonnegative
    integer or infinity) is the Lie algebra with basis
    `\{p_i\}_{1 \leq i \leq n} \cup \{q_i\}_{1 \leq i \leq n} \cup \{z\}`
    with the following relations:

    .. MATH::

        [p_i, q_j] = \delta_{ij} z, \quad [p_i, z] = [q_i, z] = [p_i, p_j]
        = [q_i, q_j] = 0.

    This Lie algebra is also known as the Heisenberg algebra of rank `n`.

    .. NOTE::

        The relations `[p_i, q_j] = \delta_{ij} z`, `[p_i, z] = 0`, and
        `[q_i, z] = 0` are known as canonical commutation relations. See
        :wikipedia:`Canonical_commutation_relations`.

    .. WARNING::

        The `n` in the above definition is called the "rank" of the
        Heisenberg algebra; it is not, however, a rank in any of the usual
        meanings that this word has in the theory of Lie algebras.

    INPUT:

    - ``R`` -- the base ring
    - ``n`` -- the rank of the Heisenberg algebra

    REFERENCES:

    - :wikipedia:`Heisenberg_algebra`

    EXAMPLES::

        sage: L = lie_algebras.Heisenberg(QQ, 2)
    """
    def __init__(self, R, n):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, 2)
            sage: TestSuite(L).run()
            sage: L = lie_algebras.Heisenberg(QQ, 0)  # not tested -- :issue:`18224`
            sage: TestSuite(L).run()
        """
        HeisenbergAlgebra_fd.__init__(self, n)
        names = tuple(['p%s' % i for i in range(1,n+1)]
                      + ['q%s' % i for i in range(1,n+1)]
                      + ['z'])
        LieAlgebraWithGenerators.__init__(self, R, names=names, index_set=names,
            category=LieAlgebras(R).Nilpotent().FiniteDimensional().WithBasis())
        HeisenbergAlgebra_abstract.__init__(self, names)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: lie_algebras.Heisenberg(QQ, 3)
            Heisenberg algebra of rank 3 over Rational Field
        """
        return "Heisenberg algebra of rank {} over {}".format(self._n, self.base_ring())


class InfiniteHeisenbergAlgebra(HeisenbergAlgebra_abstract, LieAlgebraWithGenerators):
    r"""
    The infinite Heisenberg algebra.

    This is the Heisenberg algebra on an infinite number of generators. In
    other words, this is the Heisenberg algebra of rank `\infty`. See
    :class:`HeisenbergAlgebra` for more information.
    """
    def __init__(self, R):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: TestSuite(L).run()
            sage: L.p(1).bracket(L.q(1)) == L.z()
            True
            sage: L.q(1).bracket(L.p(1)) == -L.z()
            True
        """
        S = cartesian_product([PositiveIntegers(), ['p','q']])
        cat = LieAlgebras(R).Nilpotent().WithBasis()
        LieAlgebraWithGenerators.__init__(self, R, index_set=S, category=cat)
        HeisenbergAlgebra_abstract.__init__(self, S)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: lie_algebras.Heisenberg(QQ, oo)
            Infinite Heisenberg algebra over Rational Field
        """
        return "Infinite Heisenberg algebra over {}".format(self.base_ring())

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L._an_element_()
            p2 + q2 - 1/2*q3 + z
        """
        c = self.base_ring().an_element()
        return self.p(2) + self.q(2) - c * self.q(3) + self.z()

    def lie_algebra_generators(self):
        """
        Return the generators of ``self`` as a Lie algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L.lie_algebra_generators()
            Lazy family (generator map(i))_{i in The Cartesian product of
                                            (Positive integers, {'p', 'q'})}
        """
        return Family(self._indices, lambda x: self.monomial(x[1] + str(x[0])),
                      name='generator map')

    def basis(self):
        """
        Return the basis of ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, oo)
            sage: L.basis()
            Lazy family (basis map(i))_{i in Disjoint union of Family ({'z'},
             The Cartesian product of (Positive integers, {'p', 'q'}))}
            sage: L.basis()['z']
            z
            sage: L.basis()[(12, 'p')]
            p12
        """
        S = cartesian_product([PositiveIntegers(), ['p', 'q']])
        I = DisjointUnionEnumeratedSets([Set(['z']), S])

        def basis_elt(x):
            if isinstance(x, str):
                return self.monomial(x)
            return self.monomial(x[1] + str(x[0]))
        return Family(I, basis_elt, name="basis map")

    def _from_fd_on_basis(self, i):
        """
        Return the monomial in ``self`` corresponding to the
        basis element indexed by ``i``, where ``i`` is a basis index for
        a *finite-dimensional* Heisenberg algebra.

        This is used for coercion.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, oo)
            sage: H._from_fd_on_basis('p2')
            p2
            sage: H._from_fd_on_basis('q3')
            q3
            sage: H._from_fd_on_basis('z')
            z
        """
        if i == 'z':
            return self.z()
        if i[0] == 'p':
            return self.p(Integer(i[1:]))
        return self.q(Integer(i[1:]))

    def _coerce_map_from_(self, H):
        """
        Return the coercion map from ``H`` to ``self`` if one exists,
        otherwise return ``None``.

        EXAMPLES::

            sage: H = lie_algebras.Heisenberg(QQ, oo)
            sage: HZ = lie_algebras.Heisenberg(ZZ, oo)
            sage: phi = H.coerce_map_from(HZ)
            sage: phi(HZ.p(3)) == H.p(3)
            True
            sage: phi(HZ.p(3)).leading_coefficient().parent()
            Rational Field
            sage: HF = lie_algebras.Heisenberg(QQ, 3, representation='matrix')
            sage: H.has_coerce_map_from(HF)
            True
            sage: H(HF.p(2))
            p2
            sage: H(HF.z())
            z
            sage: HF = lie_algebras.Heisenberg(QQ, 3)
            sage: H.has_coerce_map_from(HF)
            True
            sage: H(HF.p(2))
            p2
            sage: H(HF.z())
            z
        """
        if isinstance(H, HeisenbergAlgebra_fd):
            if self.base_ring().has_coerce_map_from(H.base_ring()):
                return H.module_morphism(self._from_fd_on_basis, codomain=self)
            return None  # Otherwise no coercion
        if isinstance(H, InfiniteHeisenbergAlgebra):
            if self.base_ring().has_coerce_map_from(H.base_ring()):
                return lambda C,x: self._from_dict(x._monomial_coefficients, coerce=True)
            return None  # Otherwise no coercion
        return super()._coerce_map_from_(H)


#######################################################
# Finite rank Heisenberg algebra using matrices

class HeisenbergAlgebra_matrix(HeisenbergAlgebra_fd, LieAlgebraFromAssociative):
    r"""
    A Heisenberg algebra represented using matrices.

    The `n`-th Heisenberg algebra over `R` is a Lie algebra which is
    defined as the Lie algebra of the `(n+2) \times (n+2)`-matrices:

    .. MATH::

        \begin{bmatrix}
        0 & p^T & k \\
        0 & 0_n & q \\
        0 & 0 & 0
        \end{bmatrix}

    where `p, q \in R^n` and `0_n` in the `n \times n` zero matrix. It has
    a basis consisting of

    .. MATH::

        \begin{aligned}
        p_i & = \begin{bmatrix}
        0 & e_i^T & 0 \\
        0 & 0_n & 0 \\
        0 & 0 & 0
        \end{bmatrix} \qquad \text{for } 1 \leq i \leq n ,
        \\ q_i & = \begin{bmatrix}
        0 & 0 & 0 \\
        0 & 0_n & e_i \\
        0 & 0 & 0
        \end{bmatrix} \qquad \text{for } 1 \leq i \leq n ,
        \\ z & = \begin{bmatrix}
        0 & 0 & 1 \\
        0 & 0_n & 0 \\
        0 & 0 & 0
        \end{bmatrix},
        \end{aligned}

    where `\{e_i\}` is the standard basis of `R^n`. In other words, it has
    the basis `(p_1, p_2, \ldots, p_n, q_1, q_2, \ldots, q_n, z)`, where
    `p_i = E_{1, i+1}`, `q_i = E_{i+1, n+2}` and `z = E_{1, n+2}` are
    elementary matrices.

    This Lie algebra is isomorphic to the `n`-th Heisenberg algebra
    constructed in :class:`HeisenbergAlgebra`; the bases correspond to
    each other.

    INPUT:

    - ``R`` -- the base ring
    - ``n`` -- the nonnegative integer `n`

    EXAMPLES::

        sage: L = lie_algebras.Heisenberg(QQ, 1, representation='matrix')
        sage: p = L.p(1)
        sage: q = L.q(1)
        sage: z = L.bracket(p, q); z
        [0 0 1]
        [0 0 0]
        [0 0 0]
        sage: z == L.z()
        True
        sage: L.dimension()
        3

        sage: L = lie_algebras.Heisenberg(QQ, 2, representation='matrix')
        sage: sorted(dict(L.basis()).items())
        [(
              [0 1 0 0]
              [0 0 0 0]
              [0 0 0 0]
        'p1', [0 0 0 0]
        ),
         (
              [0 0 1 0]
              [0 0 0 0]
              [0 0 0 0]
        'p2', [0 0 0 0]
        ),
         (
              [0 0 0 0]
              [0 0 0 1]
              [0 0 0 0]
        'q1', [0 0 0 0]
        ),
         (
              [0 0 0 0]
              [0 0 0 0]
              [0 0 0 1]
        'q2', [0 0 0 0]
        ),
         (
             [0 0 0 1]
             [0 0 0 0]
             [0 0 0 0]
        'z', [0 0 0 0]
        )]

        sage: L = lie_algebras.Heisenberg(QQ, 0, representation='matrix')
        sage: sorted(dict(L.basis()).items())
        [(
             [0 1]
        'z', [0 0]
        )]
        sage: L.gens()
        (
        [0 1]
        [0 0]
        )
        sage: L.lie_algebra_generators()
        Finite family {'z': [0 1]
        [0 0]}
    """
    def __init__(self, R, n):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, 2, representation='matrix')
            sage: TestSuite(L).run()
        """
        HeisenbergAlgebra_fd.__init__(self, n)
        MS = MatrixSpace(R, n+2, sparse=True)
        one = R.one()
        p = tuple(MS({(0,i): one}) for i in range(1, n+1))
        q = tuple(MS({(i,n+1): one}) for i in range(1, n+1))
        z = (MS({(0,n+1): one}),)
        names = tuple('p%s' % i for i in range(1,n+1))
        names = names + tuple('q%s' % i for i in range(1,n+1)) + ('z',)
        cat = LieAlgebras(R).Nilpotent().FiniteDimensional().WithBasis()
        LieAlgebraFromAssociative.__init__(self, MS, p + q + z, names=names,
                                           index_set=names, category=cat)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: lie_algebras.Heisenberg(QQ, 3, representation='matrix')
            Heisenberg algebra of rank 3 over Rational Field
        """
        return "Heisenberg algebra of rank {} over {}".format(self._n, self.base_ring())

    def p(self, i):
        r"""
        Return the generator `p_i` of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, 1, representation='matrix')
            sage: L.p(1)
            [0 1 0]
            [0 0 0]
            [0 0 0]
        """
        return self._gens['p%s' % i]

    def q(self, i):
        r"""
        Return the generator `q_i` of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, 1, representation='matrix')
            sage: L.q(1)
            [0 0 0]
            [0 0 1]
            [0 0 0]
        """
        return self._gens['q%s' % i]

    def z(self):
        """
        Return the basis element `z` of the Heisenberg algebra.

        The element `z` spans the center of the Heisenberg algebra.

        EXAMPLES::

            sage: L = lie_algebras.Heisenberg(QQ, 1, representation='matrix')
            sage: L.z()
            [0 0 1]
            [0 0 0]
            [0 0 0]
        """
        return self._gens['z']

    def step(self):
        r"""
        Return the nilpotency step of ``self``.

        EXAMPLES::

            sage: h = lie_algebras.Heisenberg(ZZ, 2, representation='matrix')
            sage: h.step()
            2
        """
        return Integer(2)

    class Element(LieAlgebraMatrixWrapper, LieAlgebraFromAssociative.Element):
        def monomial_coefficients(self, copy=True):
            """
            Return a dictionary whose keys are indices of basis elements in
            the support of ``self`` and whose values are the corresponding
            coefficients.

            INPUT:

            - ``copy`` -- ignored

            EXAMPLES::

                sage: L = lie_algebras.Heisenberg(QQ, 3, representation='matrix')
                sage: elt = L(Matrix(QQ, [[0, 1, 3, 0, 3], [0, 0, 0, 0, 0], [0, 0, 0, 0, -3],
                ....:                     [0, 0, 0, 0, 7], [0, 0, 0, 0, 0]]))
                sage: elt
                [ 0  1  3  0  3]
                [ 0  0  0  0  0]
                [ 0  0  0  0 -3]
                [ 0  0  0  0  7]
                [ 0  0  0  0  0]
                sage: sorted(elt.monomial_coefficients().items())
                [('p1', 1), ('p2', 3), ('q2', -3), ('q3', 7), ('z', 3)]
            """
            d = {}
            n = self.parent()._n
            for i, mon in enumerate(self.parent().basis().keys()):
                if i < n:
                    entry = self[0, i+1]
                elif i < 2 * n:
                    entry = self[i-n+1, n+1]
                else:
                    entry = self[0, n+1]
                if entry:
                    d[mon] = entry
            return d



"""
Complex multiplication for elliptic curves

This module implements the functions

- ``hilbert_class_polynomial``
- ``is_HCP``
- ``cm_j_invariants``
- ``cm_orders``
- ``discriminants_with_bounded_class_number``
- ``cm_j_invariants_and_orders``
- ``largest_fundamental_disc_with_class_number``
- ``is_cm_j_invariant``

AUTHORS:

- Robert Bradshaw
- John Cremona
- William Stein
"""

# ****************************************************************************
#       Copyright (C) 2005-2012 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer import Integer
from sage.rings.rational_field import QQ
from sage.rings.integer_ring import ZZ
from sage.rings.integer_ring import IntegerRing

from sage.misc.cachefunc import cached_function
from sage.rings.number_field.number_field_element_base import NumberFieldElement_base


@cached_function
def hilbert_class_polynomial(D, algorithm=None):
    r"""
    Return the Hilbert class polynomial for discriminant `D`.

    INPUT:

    - ``D`` -- negative integer congruent to 0 or 1 modulo 4

    - ``algorithm`` -- string (default: ``None``)

    OUTPUT:

    (integer polynomial) The Hilbert class polynomial for the
    discriminant `D`.

    ALGORITHM:

    - If ``algorithm`` = "arb" (default): Use FLINT's implementation inherited
      from Arb which uses complex interval arithmetic.

    - If ``algorithm`` = "sage": Use complex approximations to the roots.

    - If ``algorithm`` = "magma": Call the appropriate Magma function (if available).

    AUTHORS:

    - Sage implementation originally by Eduardo Ocampo Alvarez and
      AndreyTimofeev

    - Sage implementation corrected by John Cremona (using corrected precision bounds from Andreas Enge)

    - Magma implementation by David Kohel

    EXAMPLES::

        sage: # needs sage.libs.flint
        sage: hilbert_class_polynomial(-4)
        x - 1728
        sage: hilbert_class_polynomial(-7)
        x + 3375
        sage: hilbert_class_polynomial(-23)
        x^3 + 3491750*x^2 - 5151296875*x + 12771880859375
        sage: hilbert_class_polynomial(-37*4)
        x^2 - 39660183801072000*x - 7898242515936467904000000
        sage: hilbert_class_polynomial(-37*4, algorithm='magma') # optional - magma
        x^2 - 39660183801072000*x - 7898242515936467904000000
        sage: hilbert_class_polynomial(-163)
        x + 262537412640768000
        sage: hilbert_class_polynomial(-163, algorithm='sage')
        x + 262537412640768000
        sage: hilbert_class_polynomial(-163, algorithm='magma') # optional - magma
        x + 262537412640768000

    TESTS::

        sage: all(hilbert_class_polynomial(d, algorithm='arb') ==
        ....:      hilbert_class_polynomial(d, algorithm='sage')
        ....:        for d in range(-1,-100,-1) if d % 4 in [0, 1])
        True
    """
    if algorithm is None:
        algorithm = "arb"

    D = Integer(D)
    if D >= 0:
        raise ValueError("D (=%s) must be negative" % D)
    if (D % 4) not in [0, 1]:
        raise ValueError("D (=%s) must be a discriminant" % D)

    if algorithm == "arb":
        import sage.libs.arb.arith
        return sage.libs.arb.arith.hilbert_class_polynomial(D)

    if algorithm == "magma":
        from sage.interfaces.magma import magma
        magma.eval("R<x> := PolynomialRing(IntegerRing())")
        f = str(magma.eval("HilbertClassPolynomial(%s)" % D))
        return IntegerRing()['x'](f)

    if algorithm != "sage":
        raise ValueError("%s is not a valid algorithm" % algorithm)

    from sage.quadratic_forms.binary_qf import BinaryQF_reduced_representatives
    from sage.rings.real_mpfr import RR
    from sage.rings.complex_mpfr import ComplexField
    from sage.functions.special import elliptic_j

    # get all primitive reduced quadratic forms, (necessary to exclude
    # imprimitive forms when D is not a fundamental discriminant):

    rqf = BinaryQF_reduced_representatives(D, primitive_only=True)

    # compute needed precision
    #
    # NB: [https://arxiv.org/abs/0802.0979v1], quoting Enge (2006), is
    # incorrect.  Enge writes (2009-04-20 email to John Cremona) "The
    # source is my paper on class polynomials
    # [https://hal.inria.fr/inria-00001040] It was pointed out to me by
    # the referee after ANTS that the constant given there was
    # wrong. The final version contains a corrected constant on p.7
    # which is consistent with your example. It says:

    # "The logarithm of the absolute value of the coefficient in front
    # of X^j is bounded above by
    #
    # log (2*k_2) * h + pi * sqrt(|D|) * sum (1/A_i)
    #
    # independently of j", where k_2 \approx 10.163.

    h = len(rqf) # class number
    c1 = 3.05682737291380 # log(2*10.63)
    c2 = sum([1/RR(qf[0]) for qf in rqf], RR(0))
    prec = c2 * RR(3.142) * RR(D).abs().sqrt() + h * c1  # bound on log
    prec = prec * 1.45   # bound on log_2 (1/log(2) = 1.44..)
    prec = 10 + prec.ceil()  # allow for rounding error

    # set appropriate precision for further computing

    Dsqrt = D.sqrt(prec=prec)
    R = ComplexField(prec)['t']
    t = R.gen()
    pol = R(1)
    for qf in rqf:
        a, b, c = list(qf)
        tau = (b + Dsqrt) / (a << 1)
        pol *= (t - elliptic_j(tau))

    coeffs = [cof.real().round() for cof in pol.coefficients(sparse=False)]
    return IntegerRing()['x'](coeffs)


def is_HCP(f, check_monic_irreducible=True):
    r"""
    Determine whether a polynomial is a Hilbert Class Polynomial.

    INPUT:

    - ``f`` -- a polynomial in `\ZZ[X]`
    - ``check_monic_irreducible`` -- boolean (default: ``True``); if ``True``,
      check that ``f`` is a monic, irreducible, integer polynomial

    OUTPUT:

    (integer) -- either `D` if ``f`` is the Hilbert Class Polynomial
    `H_D` for discriminant `D`, or `0` if not an HCP.

    ALGORITHM:

    Cremona and Sutherland: Algorithm 2 of [CreSuth2023]_.

    EXAMPLES:

    Even for large degrees this is fast.  We test the largest
    discriminant of class number 100, for which the HCP has coefficients
    with thousands of digits::

        sage: from sage.schemes.elliptic_curves.cm import is_HCP
        sage: D = -1856563
        sage: D.class_number()                                                          # needs sage.libs.pari
        100

        sage: # needs sage.libs.flint
        sage: H = hilbert_class_polynomial(D)
        sage: H.degree()
        100
        sage: max(H).ndigits()
        2774
        sage: is_HCP(H)
        -1856563

    Testing polynomials which are not HCPs is faster::

        sage: is_HCP(H+1)                                                               # needs sage.libs.flint
        0


    TESTS::

        sage: # needs sage.libs.flint
        sage: from sage.schemes.elliptic_curves.cm import is_HCP
        sage: all(is_HCP(hilbert_class_polynomial(D)) == D
        ....:     for D in srange(-4,-100,-1) if D.is_discriminant())
        True
        sage: all(not is_HCP(hilbert_class_polynomial(D) + 1)
        ....:     for D in srange(-4,-100,-1) if D.is_discriminant())
        True

    Ensure that :issue:`37471` is fixed::

        sage: from sage.schemes.elliptic_curves.cm import is_HCP
        sage: set_random_seed(297388353221545796156853787333338705098)
        sage: is_HCP(hilbert_class_polynomial(-55))
        -55
    """
    zero = ZZ(0)
    # optional check that input is monic and irreducible
    if check_monic_irreducible:
        try:
            if not (all(c in ZZ for c in f) and f.is_monic()):
                return zero
            f = f.change_ring(ZZ)
        except AttributeError:
            return zero

    from sage.rings.real_mpfr import RR
    from sage.rings.finite_rings.finite_field_constructor import GF

    h = f.degree()
    h2list = [d for d in h.divisors()
              if (d-h) % 2 == 0 and d.prime_to_m_part(2) == 1]
    pmin = 33 * (h**2 * (RR(h+2).log().log()+2)**2).ceil()
    # Guarantees 4*p > |D| for fundamental D under GRH
    p = pmin-1
    n = 0
    from sage.arith.misc import next_prime
    from sage.schemes.elliptic_curves.constructor import EllipticCurve

    while True:
        p = next_prime(p)
        n += 1
        fp = f.change_ring(GF(p))
        # Compute X^p-X mod fp
        z = fp.parent().gen()
        r = pow(z, p, fp) - z
        r = r.gcd(fp)
        d = r.degree()  # number of roots mod p
        if d == 0:
            continue
        if not r.is_squarefree():
            continue
        if d < h and d not in h2list:
            return zero
        jp = r.any_root(degree=1, assume_squarefree=True, assume_equal_deg=True)
        E = EllipticCurve(j=jp)
        if E.is_supersingular():
            continue
        try:
            D = E.endomorphism_discriminant_from_class_number(h)
        except ValueError:
            return zero
        return D if f == hilbert_class_polynomial(D) else zero


def OrderClassNumber(D0, h0, f):
    r"""
    Return the class number h(f**2 * D0), given h(D0)=h0.

    INPUT:

    - ``D0`` -- integer; a negative fundamental discriminant
    - ``h0`` -- integer; the class number of the (maximal) imaginary quadratic order of discriminant ``D0``
    - ``f`` -- positive integer

    OUTPUT:

    (integer) the class number of the imaginary quadratic order of discriminant ``D0*f**2``

    ALGORITHM:

    We use the formula for the class number of the order `\mathcal{O}_{D}` in terms of the class number of the
     maximal order  `\mathcal{O}_{D_0}`; see [Cox1989]_ Theorem 7.24:

    .. MATH::

        h(D) = \frac{h(D_0)f}{[\mathcal{O}_{D_0}^\times:\mathcal{O}_{D}^\times]}\prod_{p\,|\,f}\left(1-\left(\frac{D_0}{p}\right)\frac{1}{p}\right)

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: from sage.schemes.elliptic_curves.cm import OrderClassNumber
        sage: D0 = -4
        sage: h = D0.class_number()
        sage: [OrderClassNumber(D0,h,f) for f in srange(1,20)]
        [1, 1, 2, 2, 2, 4, 4, 4, 6, 4, 6, 8, 6, 8, 8, 8, 8, 12, 10]
        sage: all(OrderClassNumber(D0,h,f) == (D0*f**2).class_number() for f in srange(1,20))
        True
    """
    if not D0.is_fundamental_discriminant():
        raise ValueError("{} is not a fundamental discriminant".format(D0))
    if f <= 0:
        raise ValueError("{} is not a positive integer".format(f))
    if f == 1:
        return h0
    ps = f.prime_divisors()
    from sage.misc.misc_c import prod
    from sage.arith.misc import kronecker as kronecker_symbol
    n = (f // prod(ps)) * prod(p - kronecker_symbol(D0, p) for p in ps)
    if D0 == -3:
        # assert h0 == 1 and n % 3 == 0
        return n // 3
    if D0 == -4:
        # assert h0 == 1 and n % 2 == 0
        return n // 2
    return n * h0


@cached_function
def cm_j_invariants(K, proof=None):
    r"""
    Return a list of all CM `j`-invariants in the field `K`.

    INPUT:

    - ``K`` -- a number field
    - ``proof`` -- (default: proof.number_field())

    OUTPUT:

    (list) -- A list of CM `j`-invariants in the field `K`.

    EXAMPLES::

        sage: cm_j_invariants(QQ)
        [-262537412640768000, -147197952000, -884736000, -12288000, -884736,
         -32768, -3375, 0, 1728, 8000, 54000, 287496, 16581375]

    Over imaginary quadratic fields there are no more than over `QQ`::

        sage: cm_j_invariants(QuadraticField(-1, 'i'))                                  # needs sage.rings.number_field
        [-262537412640768000, -147197952000, -884736000, -12288000, -884736,
         -32768, -3375, 0, 1728, 8000, 54000, 287496, 16581375]

    Over real quadratic fields there may be more, for example::

        sage: len(cm_j_invariants(QuadraticField(5, 'a')))                              # needs sage.rings.number_field
        31

    Over number fields K of many higher degrees this also works::

        sage: # needs sage.rings.number_field
        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^3 - 2)
        sage: cm_j_invariants(K)
        [-262537412640768000, -147197952000, -884736000, -884736, -32768,
         8000, -3375, 16581375, 1728, 287496, 0, 54000, -12288000,
         31710790944000*a^2 + 39953093016000*a + 50337742902000]
        sage: K.<a> = NumberField(x^4 - 2)
        sage: len(cm_j_invariants(K))
        23
    """
    return sorted(j for D, f, j in cm_j_invariants_and_orders(K, proof=proof))


@cached_function
def cm_j_invariants_and_orders(K, proof=None):
    r"""
    Return a list of all CM `j`-invariants in the field `K`, together with the associated orders.

    INPUT:

    - ``K`` -- a number field
    - ``proof`` -- (default: proof.number_field())

    OUTPUT:

    A list of 3-tuples `(D,f,j)` where `j` is a CM `j`-invariant in `K` with
    quadratic fundamental discriminant `D` and conductor `f`.

    EXAMPLES::

        sage: cm_j_invariants_and_orders(QQ)
        [(-3, 3, -12288000), (-3, 2, 54000), (-3, 1, 0), (-4, 2, 287496), (-4, 1, 1728),
         (-7, 2, 16581375), (-7, 1, -3375), (-8, 1, 8000), (-11, 1, -32768),
         (-19, 1, -884736), (-43, 1, -884736000), (-67, 1, -147197952000),
         (-163, 1, -262537412640768000)]

    Over an imaginary quadratic field there are no more than over `QQ`::

        sage: cm_j_invariants_and_orders(QuadraticField(-1, 'i'))                       # needs sage.rings.number_field
        [(-163, 1, -262537412640768000), (-67, 1, -147197952000),
         (-43, 1, -884736000), (-19, 1, -884736), (-11, 1, -32768),
         (-8, 1, 8000), (-7, 1, -3375), (-7, 2, 16581375), (-4, 1, 1728),
         (-4, 2, 287496), (-3, 1, 0), (-3, 2, 54000), (-3, 3, -12288000)]

    Over real quadratic fields there may be more::

        sage: v = cm_j_invariants_and_orders(QuadraticField(5,'a')); len(v)             # needs sage.rings.number_field
        31
        sage: [(D, f) for D, f, j in v if j not in QQ]                                  # needs sage.rings.number_field
        [(-235, 1), (-235, 1), (-115, 1), (-115, 1), (-40, 1), (-40, 1),
         (-35, 1), (-35, 1), (-20, 1), (-20, 1), (-15, 1), (-15, 1), (-15, 2),
         (-15, 2), (-4, 5), (-4, 5), (-3, 5), (-3, 5)]

    Over number fields K of many higher degrees this also works::

        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^3 - 2)                                              # needs sage.rings.number_field
        sage: cm_j_invariants_and_orders(K)                                             # needs sage.rings.number_field
        [(-163, 1, -262537412640768000), (-67, 1, -147197952000),
         (-43, 1, -884736000), (-19, 1, -884736), (-11, 1, -32768),
         (-8, 1, 8000), (-7, 1, -3375), (-7, 2, 16581375), (-4, 1, 1728),
         (-4, 2, 287496), (-3, 1, 0), (-3, 2, 54000), (-3, 3, -12288000),
         (-3, 6, 31710790944000*a^2 + 39953093016000*a + 50337742902000)]
    """
    if K == QQ:
        return [(ZZ(d), ZZ(f), ZZ(j)) for d, f, j in [
            (-3, 3, -12288000),
            (-3, 2, 54000),
            (-3, 1, 0),
            (-4, 2, 287496),
            (-4, 1, 1728),
            (-7, 2, 16581375),
            (-7, 1, -3375),
            (-8, 1, 8000),
            (-11, 1, -32768),
            (-19, 1, -884736),
            (-43, 1, -884736000),
            (-67, 1, -147197952000),
            (-163, 1, -262537412640768000)]]

    # Get the list of CM orders that could possibly have Hilbert class
    # polynomial F(x) with a root in K.  If F(x) has a root alpha in K,
    # then F is the minimal polynomial of alpha in K, so the degree of
    # F(x) divides [K:QQ].
    n = K.absolute_degree()
    T = discriminants_with_bounded_class_number(n, proof=proof)
    dlist = sorted(sum((Dflist for h,Dflist in T.items() if h.divides(n)), []))

    return [(D, f, j) for D, f in dlist
            for j in hilbert_class_polynomial(D*f*f).roots(K, multiplicities=False)]


@cached_function
def cm_orders(h, proof=None):
    """
    Return a list of all pairs `(D,f)` where there is a CM order of
    discriminant `D f^2` with class number h, with `D` a fundamental
    discriminant.

    INPUT:

    - ``h`` -- positive integer
    - ``proof`` -- (default: proof.number_field())

    OUTPUT: list of 2-tuples `(D,f)` sorted lexicographically by `(|D|, f)`

    EXAMPLES::

        sage: cm_orders(0)
        []
        sage: v = cm_orders(1); v
        [(-3, 1), (-3, 2), (-3, 3), (-4, 1), (-4, 2), (-7, 1), (-7, 2), (-8, 1),
         (-11, 1), (-19, 1), (-43, 1), (-67, 1), (-163, 1)]
        sage: type(v[0][0]), type(v[0][1])
        (<... 'sage.rings.integer.Integer'>, <... 'sage.rings.integer.Integer'>)
        sage: # needs sage.libs.pari
        sage: v = cm_orders(2); v
         [(-3, 4), (-3, 5), (-3, 7), (-4, 3), (-4, 4), (-4, 5), (-7, 4), (-8, 2),
          (-8, 3), (-11, 3), (-15, 1), (-15, 2), (-20, 1), (-24, 1), (-35, 1),
          (-40, 1), (-51, 1), (-52, 1), (-88, 1), (-91, 1), (-115, 1), (-123, 1),
          (-148, 1), (-187, 1), (-232, 1), (-235, 1), (-267, 1), (-403, 1), (-427, 1)]
        sage: len(v)
        29
        sage: set([hilbert_class_polynomial(D*f^2).degree() for D,f in v])
        {2}

    Any degree up to 100 is implemented, but may be slow::

        sage: # needs sage.libs.pari
        sage: cm_orders(3)
        [(-3, 6), (-3, 9), (-11, 2), (-19, 2), (-23, 1), (-23, 2), (-31, 1), (-31, 2),
         (-43, 2), (-59, 1), (-67, 2), (-83, 1), (-107, 1), (-139, 1), (-163, 2),
         (-211, 1), (-283, 1), (-307, 1), (-331, 1), (-379, 1), (-499, 1), (-547, 1),
         (-643, 1), (-883, 1), (-907, 1)]
        sage: len(cm_orders(4))
        84
    """
    h = Integer(h)
    if h <= 0:
        # trivial case
        return []

    if h in hDf_dict:
        return hDf_dict[h]

    # Get all discriminants for all class numbers up to h (which will
    # be stored in hDf_dict), and return just those with class number h.
    return discriminants_with_bounded_class_number(h, proof=proof)[h]

# Table from Mark Watkins paper "Class numbers of imaginary quadratic fields".

# WAS extracted this by cutting/pasting from the pdf, and running this program:
# z = {}
# for X in open('/Users/wstein/tmp/a.txt').readlines():
#    if X.strip():
#        v = [int(a) for a in X.split()]
#        for i in range(5):
#            z[v[3*i]]=(v[3*i+2], v[3*i+1])

# The keys are integers 1--100 and the value for each h is (|D|,n)
# where |D| is the largest absolute discriminant of an imaginary
# quadratic field with class number h, and n is the number of such
# fields.  These are all *unconditional* (not dependent on GRH).


watkins_table = {1: (163, 9), 2: (427, 18), 3: (907, 16), 4: (1555, 54), 5: (2683, 25),
                 6: (3763, 51), 7: (5923, 31), 8: (6307, 131), 9: (10627, 34), 10:
                 (13843, 87), 11: (15667, 41), 12: (17803, 206), 13: (20563, 37), 14:
                 (30067, 95), 15: (34483, 68), 16: (31243, 322), 17: (37123, 45), 18:
                 (48427, 150), 19: (38707, 47), 20: (58507, 350), 21: (61483, 85), 22:
                 (85507, 139), 23: (90787, 68), 24: (111763, 511), 25: (93307, 95), 26:
                 (103027, 190), 27: (103387, 93), 28: (126043, 457), 29: (166147, 83),
                 30: (134467, 255), 31: (133387, 73), 32: (164803, 708), 33: (222643, 101),
                 34: (189883, 219), 35: (210907, 103), 36: (217627, 668), 37:
                 (158923, 85), 38: (289963, 237), 39: (253507, 115), 40: (260947, 912),
                 41: (296587, 109), 42: (280267, 339), 43: (300787, 106), 44: (319867, 691),
                 45: (308323, 154), 46: (462883, 268), 47: (375523, 107), 48:
                 (335203, 1365), 49: (393187, 132), 50: (389467, 345), 51: (546067, 159),
                 52: (439147, 770), 53: (425107, 114), 54: (532123, 427), 55: (452083,163),
                 56: (494323, 1205), 57: (615883, 179), 58: (586987, 291),
                 59:(474307, 128), 60: (662803, 1302), 61: (606643, 132), 62: (647707, 323),
                 63: (991027, 216), 64: (693067, 1672), 65: (703123, 164), 66: (958483, 530),
                 67: (652723, 120), 68: (819163, 976), 69: (888427, 209), 70:(811507, 560),
                 71: (909547, 150), 72: (947923, 1930), 73: (886867, 119),
                 74: (951043, 407), 75: (916507, 237), 76: (1086187, 1075), 77: (1242763, 216),
                 78: (1004347, 561), 79: (1333963, 175), 80: (1165483, 2277), 81: (1030723, 228),
                 82: (1446547, 402), 83: (1074907, 150), 84: (1225387,1715),
                 85: (1285747, 221), 86: (1534723, 472), 87: (1261747, 222),
                 88:(1265587, 1905), 89: (1429387, 192), 90: (1548523, 801),
                 91: (1391083,214), 92: (1452067, 1248), 93: (1475203, 262), 94: (1587763, 509),
                 95:(1659067, 241), 96: (1684027, 3283), 97: (1842523, 185), 98: (2383747,580),
                 99: (1480627, 289), 100: (1856563, 1736)}

# Table from Janis Klaise [Klaise2012]_

# Extracted by converting pdf to text via pdf2ps and ps2txt, cutting/pasting
# and running this code:

# klaise_table = {}
# for X in open('klaise_table.txt').readlines():
#     if X.strip():
#         v = [int(a) for a in X.split()]
#         for i in range(4):
#             klaise_table[v[3*i]]=(v[3*i+2], v[3*i+1])

# The keys are integers 1--100 and the value for each h is (|D|,n)
# where |D| is the largest discriminant of an imaginary quadratic
# order with class number h, and n is the number of such orders.
# These are all *unconditional* (not dependent on GRH).

klaise_table = {1: (163, 13), 2: (427, 29), 3: (907, 25), 4: (1555, 84), 5: (2683, 29), 6: (4075, 101),
                7: (5923, 38), 8: (7987, 208), 9: (10627, 55), 10: (13843, 123), 11: (15667, 46),
                12: (19723, 379), 13: (20563, 43), 14: (30067, 134), 15: (34483, 95), 16: (35275, 531),
                17: (37123, 50), 18: (48427, 291), 19: (38707, 59), 20: (58843, 502), 21: (61483, 118),
                22: (85507, 184), 23: (90787, 78), 24: (111763, 1042), 25: (93307, 101), 26: (103027, 227),
                27: (103387, 136), 28: (126043, 623), 29: (166147, 94), 30: (137083, 473), 31: (133387, 83),
                32: (164803, 1231), 33: (222643, 158), 34: (189883, 262), 35: (210907, 111), 36: (217627, 1306),
                37: (158923, 96), 38: (289963, 284), 39: (253507, 162), 40: (274003, 1418), 41: (296587, 125),
                42: (301387, 596), 43: (300787, 123), 44: (319867, 911), 45: (308323, 231), 46: (462883, 330),
                47: (375523, 117), 48: (335203, 2895), 49: (393187, 146), 50: (389467, 445), 51: (546067, 217),
                52: (457867, 1006), 53: (425107, 130), 54: (532123, 812), 55: (452083, 177), 56: (494323, 1812),
                57: (615883, 237), 58: (586987, 361), 59: (474307, 144), 60: (662803, 2361), 61: (606643, 149),
                62: (647707, 386), 63: (991027, 311), 64: (693067, 2919), 65: (703123, 192), 66: (958483, 861),
                67: (652723, 145), 68: (819163, 1228), 69: (888427, 292), 70: (821683, 704), 71: (909547, 176),
                72: (947923, 4059), 73: (886867, 137), 74: (951043, 474), 75: (916507, 353), 76: (1086187, 1384),
                77: (1242763, 236), 78: (1004347, 925), 79: (1333963, 200), 80: (1165483, 3856), 81: (1030723, 339),
                82: (1446547, 487), 83: (1074907, 174), 84: (1225387, 2998), 85: (1285747, 246), 86: (1534723, 555),
                87: (1261747, 313), 88: (1265587, 2771), 89: (1429387, 206), 90: (1548523, 1516), 91: (1391083, 249),
                92: (1452067, 1591), 93: (1475203, 354), 94: (1587763, 600), 95: (1659067, 273), 96: (1684027, 7276),
                97: (1842523, 208), 98: (2383747, 710), 99: (1480627, 396), 100: (1856563, 2311)}


def largest_fundamental_disc_with_class_number(h):
    r"""
    Return largest absolute value of any fundamental negative discriminant with
    class number `h`, and the number of fundamental negative discriminants with
    that class number.  This is known (unconditionally) for `h` up to 100,
    by work of Mark Watkins ([Watkins2004]_).

    .. NOTE::

        The class number of a fundamental negative discriminant `D` is
        the same as the class number of the imaginary quadratic field
        `\QQ(\sqrt{D})`, so this function gives the number of such
        fields of each class number `h\le100`.  It is easy to extend
        this to larger class number conditional on the GRH, but much
        harder to obtain unconditional results.

    INPUT:

    - ``h`` -- integer

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cm import largest_fundamental_disc_with_class_number
        sage: largest_fundamental_disc_with_class_number(0)
        (0, 0)
        sage: largest_fundamental_disc_with_class_number(1)
        (163, 9)
        sage: largest_fundamental_disc_with_class_number(2)
        (427, 18)
        sage: largest_fundamental_disc_with_class_number(10)
        (13843, 87)
        sage: largest_fundamental_disc_with_class_number(100)
        (1856563, 1736)
        sage: largest_fundamental_disc_with_class_number(101)
        Traceback (most recent call last):
        ...
        NotImplementedError: largest fundamental discriminant not available for class number 101
    """
    h = Integer(h)
    if h <= 0:
        return Integer(0), Integer(0)
    try:
        B, c = watkins_table[h]
        return (Integer(B), Integer(c))
    except KeyError:
        raise NotImplementedError("largest fundamental discriminant not available for class number %s" % h)


def largest_disc_with_class_number(h):
    r"""
    Return largest absolute value of any negative discriminant with
    class number `h`, and the number of fundamental negative
    discriminants with that class number.  This is known
    (unconditionally) for `h` up to 100, by work of Mark Watkins
    [Watkins2004]_ for fundamental discriminants, extended to all
    discriminants of class number `h\le100` by Klaise [Klaise2012]_.

    .. NOTE::

        The class number of a negative discriminant `D` is
        the same as the class number of the unique imaginary quadratic order
        of discriminant `D`, so this function gives the number of such
        orders of each class number `h\le100`.  It is easy to extend
        this to larger class number conditional on the GRH, but much
        harder to obtain unconditional results.

    INPUT:

    - ``h`` -- integer

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cm import largest_disc_with_class_number
        sage: largest_disc_with_class_number(0)
        (0, 0)
        sage: largest_disc_with_class_number(1)
        (163, 13)
        sage: largest_disc_with_class_number(2)
        (427, 29)
        sage: largest_disc_with_class_number(10)
        (13843, 123)
        sage: largest_disc_with_class_number(100)
        (1856563, 2311)
        sage: largest_disc_with_class_number(101)
        Traceback (most recent call last):
        ...
        NotImplementedError: largest discriminant not available for class number 101

    For most `h\le100`, the largest fundamental discriminant with
    class number `h` is also the largest discriminant, but this is not
    the case for some `h`::

        sage: from sage.schemes.elliptic_curves.cm import largest_disc_with_class_number, largest_fundamental_disc_with_class_number
        sage: [h for h in range(1,101) if largest_disc_with_class_number(h)[0] != largest_fundamental_disc_with_class_number(h)[0]]
        [6, 8, 12, 16, 20, 30, 40, 42, 52, 70]
        sage: largest_fundamental_disc_with_class_number(6)
        (3763, 51)
        sage: largest_disc_with_class_number(6)
        (4075, 101)
    """
    h = Integer(h)
    if h <= 0:
        return Integer(0), Integer(0)
    try:
        B, c = klaise_table[h]
        return (Integer(B), Integer(c))
    except KeyError:
        raise NotImplementedError("largest discriminant not available for class number %s" % h)

# This dict has class numbers h as keys, the value at h is a complete
# list of pairs (D0,f) such that D=D0*f**2 has class number h.  We
# initialise it with h=1 only; other values will be added by calls to
# discriminants_with_bounded_class_number().


hDf_dict = {ZZ(1): [(ZZ(D), ZZ(h)) for D,h in
                    [(-3, 1), (-3, 2), (-3, 3), (-4, 1), (-4, 2), (-7, 1), (-7, 2),
                     (-8, 1), (-11, 1), (-19, 1), (-43, 1), (-67, 1), (-163, 1)]]}


def discriminants_with_bounded_class_number(hmax, B=None, proof=None):
    r"""Return a dictionary with keys class numbers `h\le hmax` and values the
    list of all pairs `(D_0, f)`, with `D_0<0` a fundamental discriminant such
    that `D=D_0f^2` has class number `h`.  If the optional bound `B` is given,
    return only those pairs with `|D| \le B`.

    INPUT:

    - ``hmax`` -- integer
    - ``B`` -- integer or ``None``; if ``None`` returns all pairs
    - ``proof`` -- this code calls the PARI function :pari:`qfbclassno`, so it
      could give wrong answers when ``proof``==``False`` (though only for
      discriminants greater than `2\cdot10^{10}`).  The default is
      the current value of ``proof.number_field()``.

    OUTPUT: dictionary

    .. NOTE::

       In case `B` is not given, then ``hmax`` must be at most 100; we
       use the tables from [Watkins2004]_ and [Klaise2012]_ to compute
       a `B` that captures all `h` up to `hmax`.

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: from sage.schemes.elliptic_curves.cm import discriminants_with_bounded_class_number
        sage: v = discriminants_with_bounded_class_number(3)
        sage: sorted(v)
        [1, 2, 3]
        sage: v[1]
        [(-3, 1), (-3, 2), (-3, 3), (-4, 1), (-4, 2), (-7, 1), (-7, 2), (-8, 1),
         (-11, 1), (-19, 1), (-43, 1), (-67, 1), (-163, 1)]
        sage: v[2]
        [(-3, 4), (-3, 5), (-3, 7), (-4, 3), (-4, 4), (-4, 5), (-7, 4), (-8, 2),
         (-8, 3), (-11, 3), (-15, 1), (-15, 2), (-20, 1), (-24, 1), (-35, 1), (-40, 1),
         (-51, 1), (-52, 1), (-88, 1), (-91, 1), (-115, 1), (-123, 1), (-148, 1),
         (-187, 1), (-232, 1), (-235, 1), (-267, 1), (-403, 1), (-427, 1)]
        sage: v[3]
        [(-3, 6), (-3, 9), (-11, 2), (-19, 2), (-23, 1), (-23, 2), (-31, 1), (-31, 2),
         (-43, 2), (-59, 1), (-67, 2), (-83, 1), (-107, 1), (-139, 1), (-163, 2),
         (-211, 1), (-283, 1), (-307, 1), (-331, 1), (-379, 1), (-499, 1), (-547, 1),
         (-643, 1), (-883, 1), (-907, 1)]
        sage: v = discriminants_with_bounded_class_number(8, proof=False)
        sage: sorted(len(v[h]) for h in v)
        [13, 25, 29, 29, 38, 84, 101, 208]

    Find all class numbers for discriminant up to 50::

        sage: sage.schemes.elliptic_curves.cm.discriminants_with_bounded_class_number(hmax=5, B=50)
        {1: [(-3, 1), (-3, 2), (-3, 3), (-4, 1), (-4, 2), (-7, 1), (-7, 2), (-8, 1), (-11, 1), (-19, 1), (-43, 1)], 2: [(-3, 4), (-4, 3), (-8, 2), (-15, 1), (-20, 1), (-24, 1), (-35, 1), (-40, 1)], 3: [(-11, 2), (-23, 1), (-31, 1)], 4: [(-39, 1)], 5: [(-47, 1)]}
    """
    hmax = Integer(hmax)

    # Easy case where we have already computed and cached the relevant values
    if hDf_dict and hmax <= max(hDf_dict):
        T = {h:Dflist for h,Dflist in hDf_dict.items() if h <= hmax}
        if B:
            for h in T:
                T[h] = [Df for Df in T[h] if Df[0].abs()*Df[1]**2 <= B]
        return T

    # imports that are needed only for this function
    from sage.arith.srange import xsrange
    from sage.structure.proof.proof import get_flag
    proof = get_flag(proof, 'number_field')

    if B is None:
        if hmax <= 100:
            # Determine how far we have to go by applying Watkins + Klaise's results.
            v = [largest_disc_with_class_number(h) for h in range(1, hmax+1)]
            B = max([b for b,_ in v])
            #print("Testing all discriminants up to {}".format(B))
            count = [0] + [cnt for _,cnt in v]
        else:
            raise ValueError("if hmax>100 you must specify a discriminant bound B")
    else:
        # Nothing to do -- set to None so we can use this later to know not
        # to do a double check about how many we find.
        count = None
        B = Integer(B)

    if B <= 2:
        # This is an easy special case, since there are no fundamental discriminants
        # this small.
        return {}

    # T stores the values found, to be returned.  It will also be used
    # to update the global hDf_dict *provided that* the parameter B
    # was not provided.  Note that we only reach this point if we have
    # not already computed data for at least one class number, hmax.

    # To avoid recomputing class numbers, we initialise T with the
    # global hDf_dict, only including those (D,f) for which |D|*f**2
    # <= B if B was provided:

    # h_dict caches the class number h of all discriminants previously
    # encountered; we will use the function OrderClassNumber() to
    # quickly compute the class number of non-fundamental discriminants
    # from the fundamental ones. Note that in the initialisation, the
    # keys of h_dict include nonfundamental discriminants, but we only
    # update it with fundamental ones.

    from collections import defaultdict
    T = defaultdict(set)
    h_dict = {}
    for h, Dflist in hDf_dict.items():
        for D0,f in Dflist:
            h_dict[D0*f**2] = h
        if not count:
            Dflist = [Df for Df in Dflist if Df[0].abs()*Df[1]**2 <= B]
        T[h] = set(Dflist)

    # We do not need to certify the class number from :pari:`qfbclassno` for discriminants under 2*10^10
    if B < 2*10**10:
        proof = False

    for D in xsrange(-3, -B-1, -1):
        if not D.is_discriminant():
            continue
        D0 = D.squarefree_part()
        if D0 % 4 != 1:
            D0 *= 4
        f = (D//D0).isqrt()

        # Now D0 is the fundamental discriminant and f the conductor

        if D in h_dict:
            h = h_dict[D]
        else:
            if f == 1: # D itself is fundamental
                h = D.class_number(proof)
                h_dict[D] = h
            else:
                h = OrderClassNumber(D0,h_dict[D0],f)

        # If the class number of this order is within the range, then store (D0,f)
        if h <= hmax:
            T[h].add((D0,f))

    # sort each list of (D,f) pairs by (|D|,f)

    for h in T:
        T[h] = list(T[h])
        T[h].sort(key=lambda Df: (Df[0].abs(), Df[1]))

    # count is None precisely when the user provided a value of B
    if count is not None:
        # 1. Check that we found the right number of discriminants
        for h in T:
            if len(T[h]) != count[h]:
                raise RuntimeError("number of discriminants inconsistent with Watkins's table")
        # 2. Update the global dict
        hDf_dict.update(dict(T))

    return T


@cached_function
def is_cm_j_invariant(j, algorithm='CremonaSutherland', method=None):
    r"""
    Return whether or not this is a CM `j`-invariant, and the CM discriminant if it is.

    INPUT:

    - ``j`` -- an element of a number field `K`

    - ``algorithm`` -- string (default: ``'CremonaSutherland'``); the algorithm
      used, either ``'CremonaSutherland'`` (the default, very much faster
      for all but very small degrees), ``'exhaustive'`` or ``'reduction'``

    - ``method`` -- string; deprecated name for ``algorithm``

    OUTPUT:

    A pair (bool, (d,f)) which is either (False, None) if `j` is not a
    CM j-invariant or (True, (d,f)) if `j` is the `j`-invariant of the
    imaginary quadratic order of discriminant `D=df^2` where `d` is
    the associated fundamental discriminant and `f` the index.

    ALGORITHM:

    The default algorithm used is to test whether the minimal
    polynomial of ``j`` is a Hilbert CLass Polynomial, using
    :func:`is_HCP` which implements Algorithm 2 of [CreSuth2023]_ by
    Cremona and Sutherland.

    Two older algorithms are available, both of which are much slower
    except for very small degrees.

    Method 'exhaustive' makes use of the complete and unconditionsl classification of
    all orders of class number up to 100, and hence will raise an
    error if `j` is an algebraic integer of degree greater than
    this.

    Method 'reduction' constructs an elliptic curve over the number
    field `\QQ(j)` and computes its traces of Frobenius at several
    primes of degree 1.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cm import is_cm_j_invariant
        sage: is_cm_j_invariant(0)
        (True, (-3, 1))
        sage: is_cm_j_invariant(8000)
        (True, (-8, 1))

        sage: # needs sage.rings.number_field
        sage: K.<a> = QuadraticField(5)
        sage: is_cm_j_invariant(282880*a + 632000)
        (True, (-20, 1))
        sage: x = polygen(ZZ, 'x')
        sage: K.<a> = NumberField(x^3 - 2)
        sage: is_cm_j_invariant(31710790944000*a^2 + 39953093016000*a + 50337742902000)
        (True, (-3, 6))

    An example of large degree.  This is only possible using the default algorithm::

        sage: from sage.schemes.elliptic_curves.cm import is_cm_j_invariant
        sage: D = -1856563
        sage: H = hilbert_class_polynomial(D)                                           # needs sage.libs.flint
        sage: H.degree()                                                                # needs sage.libs.flint
        100
        sage: K.<j> = NumberField(H)                                                    # needs sage.libs.flint sage.rings.number_field
        sage: is_cm_j_invariant(j)                                                      # needs sage.libs.flint sage.rings.number_field
        (True, (-1856563, 1))

    TESTS::

        sage: from sage.schemes.elliptic_curves.cm import is_cm_j_invariant
        sage: all(is_cm_j_invariant(j) == (True, (d,f)) for d,f,j in cm_j_invariants_and_orders(QQ))
        True
    """
    if method:
        if not algorithm:
            algorithm = method
        raise DeprecationWarning("'method' is deprecated, use 'algorithm instead'")

    # First we check that j is an algebraic number:
    if not isinstance(j, NumberFieldElement_base) and j not in QQ:
        raise NotImplementedError("is_cm_j_invariant() is only implemented for number field elements")

    # for j in ZZ we have a lookup-table:

    if j in ZZ:
        j = ZZ(j)
        table = {jj: (d,f) for d,f,jj in cm_j_invariants_and_orders(QQ)}
        if j in table:
            return True, table[j]
        return False, None

    # Otherwise if j is in Q then it is not integral so is not CM:

    if j in QQ:
        return False, None

    # Next we find its minimal polynomial of j:

    if j.parent().absolute_degree() == 2:
        jpol = j.absolute_minpoly() # no algorithm parameter
    else:
        jpol = j.absolute_minpoly(algorithm='pari')

    # If it does not have integer coefficients then j is not integral, hence not CM:

    if not all(c in ZZ for c in jpol):
        return False, None

    # Otherwise test whether it is a Hilbert Class Polynomial
    # (using the fact that we know that it is monic and irreducible):

    if algorithm == 'CremonaSutherland':
        D = is_HCP(jpol, check_monic_irreducible=False)
        if D:
            D0 = D.squarefree_part()
            if D0 % 4 != 1:
                D0 *= 4
            f = ZZ(D // D0).isqrt()
            return (True, (D0, f))
        return (False, None)

    h = jpol.degree()
    if algorithm in ['exhaustive', 'old']:
        if h > 100:
            raise NotImplementedError("CM data only available for class numbers up to 100")
        for d,f in cm_orders(h):
            if jpol == hilbert_class_polynomial(d*f**2):
                return (True, (d,f))
        return (False, None)

    if algorithm not in ['reduction', 'new']:
        raise ValueError("Invalid algorithm {} in is_cm_j_invariant".format(algorithm))

    # Now we use the reduction algorithm

    # If the degree h is less than the degree of j.parent() we recreate j as an element
    # of Q(j, and replace j by a clone whose parent is Q(j), if necessary:

    K = j.parent()
    if h < K.absolute_degree():
        from sage.rings.number_field.number_field import NumberField

        K = NumberField(jpol, 'j')
        j = K.gen()

    # Construct an elliptic curve with j-invariant j, with
    # integral model:

    from sage.schemes.elliptic_curves.constructor import EllipticCurve
    E = EllipticCurve(j=j).integral_model()
    D = E.discriminant()
    prime_bound = 1000  # test primes of degree 1 up to this norm
    max_primes = 20     # test at most this many primes
    num_prime = 0
    cmd = 0
    cmf = 0

    # Test primes of good reduction.  If E has CM then for half the
    # primes P we will have a_P=0, and for all other prime P the CM
    # field is Q(sqrt(a_P^2-4N(P))).  Hence if these fields are
    # different for two primes then E does not have CM.  If they are
    # all equal for the primes tested, then we have a candidate CM
    # field.  Moreover the discriminant of the endomorphism ring
    # divides all the values a_P^2-4N(P), since that is the
    # discriminant of the order containing the Frobenius at P.  So we
    # end up with a finite number (usually one) of candidate
    # discriminants to test.  Each is tested by checking that its class
    # number is h, and if so then that j is a root of its Hilbert
    # class polynomial.  In practice non CM curves will be eliminated
    # by the local test at a small number of primes (probably just 2).

    for P in K.primes_of_degree_one_iter(prime_bound):
        if num_prime > max_primes:
            if cmd: # we have a candidate CM field already
                break
            else:   # we need to try more primes
                max_primes *= 2
        if D.valuation(P) > 0: # skip bad primes
            continue
        aP = E.reduction(P).trace_of_frobenius()
        if aP == 0: # skip supersingular primes
            continue
        num_prime += 1
        DP = aP**2 - 4*P.norm()
        dP = DP.squarefree_part()
        fP = ZZ(DP//dP).isqrt()
        if cmd == 0:      # first one, so store d and f
            cmd = dP
            cmf = fP
        elif cmd != dP: # inconsistent with previous
            return (False, None)
        else:           # consistent d, so update f
            cmf = cmf.gcd(fP)

    if cmd == 0: # no conclusion, we found no degree 1 primes, revert to default algorithm
        return is_cm_j_invariant(j)

    # it looks like cm by disc cmd * f**2 where f divides cmf

    if cmd % 4 != 1:
        cmd = cmd * 4
        cmf = cmf // 2

    # Now we must check if h(cmd*f**2)==h for f|cmf; if so we check
    # whether j is a root of the associated Hilbert class polynomial.
    h0 = cmd.class_number()
    for f in cmf.divisors(): # only positive divisors
        if h != OrderClassNumber(cmd,h0,f):
            continue
        if jpol == hilbert_class_polynomial(cmd*f**2):
            return (True, (cmd, f))
    return (False, None)



from sage.rings.integer cimport Integer
from sage.rings.real_mpfr cimport RealNumber
from sage.stats.distributions.dgs cimport dgs_disc_gauss_mp_t, dgs_disc_gauss_dp_t
from sage.structure.sage_object cimport SageObject


cdef class DiscreteGaussianDistributionIntegerSampler(SageObject):
    cdef readonly RealNumber sigma
    cdef readonly RealNumber c
    cdef readonly Integer tau
    cdef readonly object algorithm
    cdef dgs_disc_gauss_mp_t *_gen_mp
    cdef dgs_disc_gauss_dp_t *_gen_dp
